import six
import collections
from .support import Status, Defaults, ResourceChanges
from .result import serializeValue, ChangeAware, Results
from .util import AutoRegisterClass, lookupClass, validateSchema, findSchemaErrors, UnfurlTaskError, toEnum, UnfurlAddingResourceError
from .eval import Ref, mapValue, RefContext
from .runtime import NodeInstance

from ruamel.yaml import YAML
yaml = YAML()

import logging
logger = logging.getLogger('unfurl')

# we want ConfigurationSpec to be standalone and easily serializable
class ConfigurationSpec(object):

  @classmethod
  def getDefaults(self):
      return dict(className=None, majorVersion=0, minorVersion='', intent=Defaults.intent,
      inputs=None, inputSchema=None, preConditions=None, postConditions=None)

  def __init__(self, name, action, className=None, majorVersion=0, minorVersion='',
      intent=Defaults.intent,
      inputs=None, inputSchema=None, preConditions=None, postConditions=None):
    assert name and className, "missing required arguments"
    self.name = name
    self.action = action
    self.className = className
    self.majorVersion = majorVersion
    self.minorVersion = minorVersion
    self.intent = intent
    self.inputs = inputs or {}
    self.inputSchema = inputSchema or {}
    self.preConditions = preConditions
    self.postConditions = postConditions

  def findInvalidateInputs(self, inputs):
    return findSchemaErrors(serializeValue(inputs), self.inputSchema)

  # XXX same for postConditions
  def findInvalidPreconditions(self, target):
    if not self.preConditions:
      return []
    # XXX this should be like a Dependency object
    expanded = serializeValue(target.attributes)
    return findSchemaErrors(expanded, self.preConditions)

  def create(self):
    # XXX2 throw clearer exception if couldn't load class
    return lookupClass(self.className)(self)

  def shouldRun(self):
    return Defaults.shouldRun

  def copy(self, **mods):
    args = self.__dict__.copy()
    args.update(mods)
    return ConfigurationSpec(**args)

  def __eq__(self, other):
    if not isinstance(other, ConfigurationSpec):
      return False
    # XXX3 add unit tests
    return (self.name == other.name and self.className == other.className
      and self.majorVersion == other.majorVersion and self.minorVersion == other.minorVersion
      and self.intent == other.intent and self.inputs == other.inputs and self.inputSchema == self.inputSchema
      and self.preConditions == other.preConditions and self.postConditions == other.postConditions)

class ConfiguratorResult(object):
  """
  If applied is True,
  the current pending configuration is set to the effective, active one
  and the previous configuration is no longer in effect.

  Modified indicates whether the underlying state of configuration,
  was changed i.e. the physically altered the system this configuration represents.

  Readystate reports the Status of the current configuration.
  """
  def __init__(self, applied, modified, readyState=None, configChanged=None, result=None):
    self.applied = applied
    self.modified = modified
    self.readyState = readyState
    self.configChanged = configChanged
    self.result = result

  def __str__(self):
    result = '' if self.result is None else str(self.result)
    return 'changes: ' + (' '.join(filter(None,[
      self.applied and 'applied',
      self.modified and 'modified',
      self.readyState and self.readyState.name,
      self.configChanged and 'config'])) or 'none') + ' ' + result

@six.add_metaclass(AutoRegisterClass)
class Configurator(object):
  def __init__(self, configurationSpec):
    self.configSpec = configurationSpec

  # yields a JobRequest, TaskRequest or a ConfiguratorResult
  def run(self, task):
    yield task.createResult(False, False)

  def dryRun(self, task):
    yield task.createResult(False, False)

  def cantRun(self, task):
    """
    Does this configurator support the requested action and parameters
    given the current state of the resource?
    (e.g. can we upgrade from the previous configuration?)

    Returns False or an error message (list or string)
    """
    return False

  def shouldRun(self, task):
    """Does this configuration need to be run?"""
    return self.configSpec.shouldRun()

  # XXX3 should be called during when checking dependencies
  # def checkConfigurationStatus(self, task):
  #   """Is this configuration still valid?"""
  #   return Status.ok

class TaskView(object):
  """
  The interface presented to configurators.
  """
  def __init__(self, manifest, configSpec, target, reason = None, dependencies=None):
    # public:
    self.configSpec = configSpec
    self.target = target
    self.reason = reason
    # XXX refcontext should include TARGET HOST etc.
    self.inputs = mapValue(self.configSpec.inputs, target)
    # private:
    self._manifest = manifest
    self.messages = []
    self._addedResources = []
    self._dependenciesChanged = False
    self.dependencies = dependencies or {}
    self._resourceChanges = ResourceChanges()

  def addMessage(self, message):
    self.messages.append(message)

  def findResource(self, name):
     return self._manifest.getRootResource().findResource(name)

  def createResult(self, applied, modified, readyState=None, configChanged=None, result=None):
    readyState = toEnum(Status, readyState)
    if applied and (not readyState or readyState == Status.notapplied):
        raise UnfurlTaskError(self, "need to set readyState if configuration was applied")

    return ConfiguratorResult(applied, modified, readyState, configChanged, result)

  # updates can be marked as dependencies (changes to dependencies changed) or required (error if changed)
  # configuration has cumulative set of changes made it to resources
  # updates update those changes
  # other configurations maybe modify those changes, triggering a configuration change
  def query(self, query, dependency=False, name=None, required=False, wantList=False, resolveExternal=True):
    # XXX refcontext should include TARGET HOST etc.
    result = Ref(query).resolve(RefContext(self.target, resolveExternal=resolveExternal), wantList)
    if dependency:
      self.addDependency(query, result, name=name, required=required, wantList=wantList)
    return result

  def addDependency(self, expr, expected=None, schema=None, name=None, required=False, wantList=False):
    getter = getattr(expr, 'asRef', None)
    if getter:
      # expr is a configuration or resource or ExternalValue
      expr = Ref(getter()).source

    dependency = Dependency(expr, expected, schema, name, required, wantList)
    self.dependencies[name or expr] = dependency
    self.dependenciesChanged = True
    return dependency

  def removeDependency(self, name):
    old = self.dependencies.pop(name, None)
    if old:
      self.dependenciesChanged = True
    return old

  def createConfigurationSpec(self, name, configSpec):
    if isinstance(configSpec, six.string_types):
      configSpec = yaml.load(configSpec)
    return self.manifest.loadConfigSpec(name, configSpec)

  def createSubTask(self, configSpec, resource=None, persist=False, required=False):
    from .job import TaskRequest
    # XXX:
    #if persist or required:
    #  expr = "::%s::.configurations::%s" % (configSpec.target, configSpec.name)
    #  self.addDependency(expr, required=required)

    if resource is None:
      resource = self.target
    return TaskRequest(configSpec, resource, persist, required)

  # # XXX how???
  # # Configurations created by subtasks are transient insofar as the are not part of the spec,
  # # but they are recorded as part of the resource's configuration state.
  # # Marking as persistent or required will create a dependency on the new configuration.
  # # XXX3 have a way to update spec attributes to trigger config updates e.g. add dns entries via attributes on a dns
  # def createSubTask(self, configSpec, persist=False, required=False):
  #   if persist or required:
  #     expr = "::%s::.configurations::%s" % (configSpec.target, configSpec.name)
  #     self.addDependency(expr, required=required)
  #   return TaskRequest(configSpec, persist, required)
  #
  # # XXX how can we explicitly associate relations with target resources etc.?
  # # through capability attributes and dependencies/relationship attributes
  def updateResources(self, resources):
    """
    Either a list or string that is parsed as YAML
    Operational state indicates if it current exists or not
    Will instantiate a new job, yield the return value to run that job right away

    .. code-block:: YAML

      - name:
        template: # name of node template
        priority: required
        dependent: boolean
        parent:
        attributes:
        status:
          readyState: ok
    """
    from .manifest import Manifest
    from .job import JobRequest
    if isinstance(resources, six.string_types):
      resources = yaml.load(resources)

    errors = []
    newResources = []
    newResourceSpecs = []
    for resourceSpec in resources:
      originalResourceSpec = resourceSpec
      try:
        rname = resourceSpec['name']
        if rname == '.self':
          existingResource = self.target
        else:
          existingResource = self.findResource(rname)
        if existingResource:
          # XXX2 if spec is defined (not just status), there should be a way to
          # indicate this should replace an existing resource or throw an error
          status = resourceSpec.get('status')
          operational = Manifest.loadStatus(status)
          if operational.localStatus:
            existingResource.localStatus = operational.localStatus
          attributes = resourceSpec.get('attributes')
          if attributes:
            for key, value in mapValue(attributes, existingResource).items():
               existingResource.attributes[key] = value
               logger.debug('setting attribute %s with %s on %s', key, value, existingResource.name)
          logger.info("updating resources %s", existingResource.name)
          continue

        resource = self.manifest.loadResource(rname, resourceSpec, parent=self.target.root)

        if resource.required or resourceSpec.get('dependent'):
          self.addDependency(resource, required=resource.required)
      except:
        errors.append(UnfurlAddingResourceError(self, originalResourceSpec))
      else:
        newResourceSpecs.append(originalResourceSpec)
        newResources.append(resource)

    if newResourceSpecs:
      self._resourceChanges.addResources(newResourceSpecs)
      self._addedResources.extend(newResources)
      logger.info("add resources %s", newResources)

      jobRequest = JobRequest(newResources, errors)
      if self.job:
        self.job.jobRequestQueue.append(jobRequest)
      return jobRequest
    return None

class Dependency(ChangeAware):
  """
  Represents a runtime dependency for a configuration.

  Dependencies are used to determine if a configuration needs re-run as follows:

  * They are dynamically created when evaluating and comparing the configuration spec's attributes with the previous
    values

  * Persistent dependencies can be created when the configurator invoke these apis: `createConfiguration`, `addResources`, `query`, `addDependency`
  """

  def __init__(self, expr, expected=None, schema=None, name=None, required=False, wantList=False):
    """
    if schema is not None, validate the result using schema
    if expected is not None, test that result equals expected
    otherwise test that result isn't empty has not changed since the last attempt
    """
    assert not (expected and schema)
    self.expr = expr

    self.expected = expected
    self.schema = schema
    self.required = required
    self.name = name
    self.wantList= wantList

  def refresh(self, config):
    if self.expected is not None:
      changeId = config.changeId
      context = RefContext(config.target, dict(val=self.expected, changeId=changeId))
      result = Ref(self.expr).resolve(context, wantList=self.wantList)
      self.expected = result

  @staticmethod
  def hasValueChanged(value, changeset):
    if isinstance(value, Results):
      return Dependency.hasValueChanged(value._attributes, changeset)
    elif isinstance(value, collections.Mapping):
      if any(Dependency.hasValueChanged(v, changeset) for v in value.values()):
        return True
    elif isinstance(value, (collections.MutableSequence, tuple)):
      if any(Dependency.hasValueChanged(v, changeset) for v in value):
        return True
    elif isinstance(value, ChangeAware):
        return value.hasChanged(changeset)
    else:
      return False

  def hasChanged(self, config):
    changeId = config.changeId
    context = RefContext(config.target, dict(val=self.expected, changeId=changeId))
    result = Ref(self.expr).resolveOne(context) #resolve(context, self.wantList)

    if self.schema:
      # result isn't as expected, something changed
      if not validateSchema(result, self.schema):
        return False
    else:
      if self.expected is not None:
        expected = mapValue(self.expected, context)
        if result != expected:
          logger.debug("hasChanged: %s != %s", result, expected)
          return True
      elif not result:
        # if expression no longer true (e.g. a resource wasn't found), then treat dependency as changed
        return True

    if self.hasValueChanged(result, config):
      return True

    return False



import sys
#from unfurl.eval import Ref # broken in 2.7

# from ansible.errors import AnsibleError

#module name is lookup name
from ansible.plugins.lookup import LookupBase
class LookupModule(LookupBase):
  def run(self, terms, variables, **kwargs):
    # resource should be current host or current config if no host
    refContext = variables['__unfurl']
    # workaround for 2.7
    Ref = sys.modules[type(variables['__unfurl']).__module__].Ref

    return list(map(lambda term: Ref(term).resolveOne(refContext), terms))



import os
import os.path
import git
from git.repo.fun import (
    is_git_dir
)
import logging
logger = logging.getLogger('unfurl')

def findGitRepo(path, isFile=True, importLoader=None):
  """
  Returns (repoURL, filePath, revision)
  RepoURL will be an empty string if it isn't a path to a git repo
  """
  # XXX if importLoader: find the repository based on the path, and check revision
  # hack for now: if ends in .git
  # XXX path can end in #revision, return it
  # see https://docs.docker.com/engine/reference/commandline/build/ for git urls like:
  # myrepo.git#mybranch, myrepo.git#pull/42/head, myrepo.git#:myfolder, myrepo.git#master:myfolder
  a, b, c = path.partition('.git/')
  if not b:
    return '', a, ''
  else:
    return a + '.git', c, ''

class Repo(object):
  # @staticmethod
  # def makeRepo(url, repotype, basedir):
  #   # XXX git or simple based on url
  #   # basedir is the project/subproject root or local-config root dependening where the import definition lives
  #   if repotype=='instance':
  #     dir = os.path.join(basedir, 'instances', 'current')
  #   else:
  #     dir = os.path.join(basedir, repotype)
  #   # XXX error if exists, else mkdirs
  #   return SimpleRepo(url, dir)

  @staticmethod
  def findGitWorkingDirs(rootDir, gitDir='.git'):
    workingDirs = {}
    for root, dirs, files in os.walk(rootDir):
      if gitDir in dirs:
        del dirs[:]  # don't visit sub directories
        repo = GitRepo(git.Repo(root))
        workingDirs[root] = (repo.url, repo)
    return workingDirs

  @staticmethod
  def createGitRepoIfExists(rootDir, gitDir='.git'):
    if is_git_dir(os.path.join(rootDir, gitDir)):
      return GitRepo(git.Repo(rootDir))
    return None

  def findPath(self, path, importLoader=None):
    base = self.workingDir
    if not base:
      return None, None, None
    repoRoot = os.path.abspath(base)
    abspath = os.path.abspath(path).rstrip('/')
    if repoRoot in abspath:
      # XXX find pinned
      # if importLoader:
      #   revision = importLoader.getRevision(self)
      # else:
      if True:
        revision = self.revision
      bare = not self.workingDir or revision != self.revision
      return abspath[len(repoRoot)+1:], revision, bare
    return None, None, None

  def isValidSpecRepo(self):
    return os.path.isfile(os.path.join(self.workingDir, 'manifest-template.yaml'))

  @classmethod
  def createWorkingDir(cls, gitUrl, localRepoPath, revision='HEAD'):
    empty_repo = git.Repo.init(localRepoPath)
    origin = empty_repo.create_remote('origin', gitUrl)
    empty_repo.create_head('master', origin.refs.master)
    empty_repo.set_tracking_branch(origin.refs.master).checkout(revision)
    return GitRepo(empty_repo)

class GitRepo(Repo):
  def __init__(self, gitrepo):
    self.repo = gitrepo
    self.url = self.workingDir or gitrepo.git_dir
    if gitrepo.remotes:
      try:
        remote = gitrepo.remotes['origin']
      except:
        remote = gitrepo.remotes[0]
      self.url = remote.url

  @property
  def workingDir(self):
    dir = self.repo.working_tree_dir
    if not dir or dir[-1] == '/':
      return dir
    else:
      return dir + '/'

  @property
  def revision(self):
    return self.repo.head.commit.hexsha

  def runCmd(self, args):
    """
    :return:
      tuple(int(status), str(stdout), str(stderr))
    """
    gitcmd = self.repo.git
    call = [gitcmd.GIT_PYTHON_GIT_EXECUTABLE]
    # add persistent git options
    call.extend(gitcmd._persistent_git_options)
    call.extend(list(args))

    return gitcmd.execute(call, with_exceptions=False, with_extended_output=True)

  def show(self, path, commitId):
    if os.path.abspath(path) and self.workingDir:
      path = path[len(self.workingDir):]
    return self.repo.git.show(commitId+':'+path)

  def checkout(self, revision=''):
    # if revision isn't specified and repo is not pinned:
    #  save the ref of current head
    self.repo.git.checkout(revision)
    logger.info("checking out '%s' at %s to %s", self.url, revision or 'HEAD', self.workingDir)
    return self.workingDir

  # XXX: def getDependentRepos()
  # XXX: def isDirty()
  # XXX: def canManage()

  def canMakeClean(self):
    for repo in self.getDependentRepos():
      if not repo.canMakeClean():
        return False
      elif repo.isDirty() and not self.canManage(repo):
        return False
    return True

  def _commitAll(self, parent=None):
    committed = []
    for repo in self.getDependentRepos():
      if repo.isDirty():
        assert self.canManage(repo)
        repo._commitAll(self)
        committed.append(repo)
    self.updateChildCommits(committed)
    self._commit()

  def getDirtyDependents(self):
    for repo in self.getDependentRepos():
      if repo.isDirty():
        yield repo

  # XXX unused.. logic is currently in yamlmanifest.commitJob()
  def commit(self):
    # before run referenced dirty repos should be committed?
    # at the very least the state of any declared repo should be saved
    # otherwise two different runs of the same commit could pull different versions
    # this is true for the spec repos also -- save in spec's manifest-template?
    repo = self.repo
    repo.index.add('*')
    # commit the manifest first so we can get a commit ref for the changerecord
    commit = repo.git.commit('')
    changeFiles = self.manifest.saveChanges(commit.hexsha)
    repo.index.add(changeFiles)
    repo.git.commit('')

  def clone(self, newPath):
    return GitRepo(self.repo.clone(newPath))

# class SimpleRepo(Repo):
#   def __init__(self, lastCommitId):
#     self.lastCommitId = int(lastCommitId or 0)
#
#   def checkout(self, commitid, useCurrent):
#     if useCurrent and commitid == self.lastCommitId:
#       return self.workingDir
#     return './revisions/{commitid}/files'
#
#   def commit(self):
#     self.lastCommitId += 1
#     # copy current workingDir to /revisions/{commitid}/files
#     return self.lastCommitId

class RevisionManager(object):
  def __init__(self, manifest, localEnv=None):
    self.manifest = manifest
    self.revisions = {manifest.specDigest: manifest}
    self.localEnv = localEnv

  def getRevision(self, change):
    digest = change['specDigest']
    commitid = change['startCommit']
    if digest in self.revisions:
      return self.revisions[digest]
    else:
      from .manifest import SnapShotManifest
      manifest = SnapShotManifest(self.manifest, commitid)
      self.revisions[digest] = manifest
      return manifest



"""
The basic operation of Unfurl is to apply the specified configuration to a resource
and record the results.

A manifest can contain a reproducible history of changes to a resource.
This history is stored in the resource definition so it doesn't
rely on git history for this. In fact a git repository isn't required at all.
But the intent is for commits in a git repo to correspond to reproducible configuration state of the system.
The git repo is also used to record or archive exact versions of each configurators applied.
"""

import pbr.version
__version__ = pbr.version.VersionInfo(__name__).version_string()

import os
import sys
vendor_dir = os.path.join(os.path.abspath(os.path.dirname(__file__)), 'vendor')
sys.path.insert(0, vendor_dir)

import logging
logging.captureWarnings(True)
_logHandler = None
def initLogging(level, logfile=None):
  rootLogger = logging.getLogger()
  rootLogger.setLevel(logging.DEBUG) # need to set this first

  if logfile:
    ch = logging.FileHandler(logfile)
    formatter = logging.Formatter(
        '[%(asctime)s] %(name)s:%(levelname)s: %(message)s')
    ch.setFormatter(formatter)
    ch.setLevel(logging.DEBUG)
    rootLogger.addHandler(ch)

  # global _logHandler so we can call initLogging multiple times
  global _logHandler
  if not _logHandler:
    _logHandler = logging.StreamHandler()
    formatter = logging.Formatter('%(name)s:%(levelname)s: %(message)s')
    _logHandler.setFormatter(formatter)
    rootLogger.addHandler(_logHandler)

  _logHandler.setLevel(level)
  return _logHandler

_logEnv = os.getenv('UNFURL_LOGGING')
if _logEnv is not None:
  initLogging(_logEnv)



import codecs
from .util import sensitive_str
from .configurator import Configurator #, Status
from .ansibleconfigurator import AnsibleConfigurator
import json
from ansible.module_utils.k8s.common import K8sAnsibleMixin

class ClusterConfigurator(Configurator):

  @staticmethod
  def _getHost(connectionConfig):
    client = K8sAnsibleMixin().get_api_client(**connectionConfig)
    return client.configuration.host

  def run(self, task):
    # just test the connection
    task.target.attributes['apiServer'] = self._getHost(task.inputs.get('connection', {}))
    yield task.createResult(True, False, "ok")

class ResourceConfigurator(AnsibleConfigurator):
  def dryRun(self, task):
    # XXX don't use print()
    print("generating playbook")
    #print(self.findPlaybook(task))
    #print(self.findPlaybook(task))
    print(json.dumps(self.findPlaybook(task), indent=4))
    yield task.createResult(False, False)

  def makeSecret(self, data):
    # base64 adds trailing \n so strip it out
    return dict(type='Opaque', apiVersion='v1', kind='Secret',
        data={k: codecs.encode(str(v).encode(), 'base64').decode().strip() for k, v in data.items()})

  def getDefinition(self, task):
    if task.target.template.isCompatibleType('unfurl.nodes.K8sNamespace'):
      return dict(apiVersion='v1', kind='Namespace')
    elif task.target.template.isCompatibleType('unfurl.nodes.k8sSecretResource'):
      return self.makeSecret(task.target.attributes.get('data', {}))
    else:
      # XXX if definition is string: parse
      # get copy so subsequent modifications dont affect the definition
      return task.target.attributes.getCopy('definition', {})

  def updateMetadata(self, definition, task):
    namespace = None
    if task.target.parent.template.isCompatibleType('unfurl.nodes.K8sNamespace'):
      namespace = task.target.parent.attributes['name']
    md = definition.setdefault('metadata', {})
    if namespace and 'namespace' not in md:
      md['namespace'] = namespace
    #else: error if namespace mismatch?

    # XXX if using target.name, convert into kube friendly dns-style name
    name = task.target.attributes.get('name', task.target.name)
    if 'name' in md and md['name'] != name:
      task.target.attributes['name'] = md['name']
    else:
      md['name'] = name

  def findPlaybook(self, task):
    definition = self.getDefinition(task)
    self.updateMetadata(definition, task)
    state = 'absent' if task.configSpec.action == 'delete' else 'present'
    connection = task.inputs.get('connection') or {}
    moduleSpec = dict(state=state, definition=definition, **connection)
    return [dict(k8s=moduleSpec)]

  def _processResult(self, task, result):
    resource = result.get('result')
    task.target.attributes['apiResource'] = resource
    data = resource and resource.get('kind') == 'Secret' and resource.get('data')
    if data:
      resource['data'] = {k: sensitive_str(v) for k, v in data.items()}

  def getResultKeys(self, task, results):
    # save first time even if it hasn't changed
    return ['result'] # also "method", "diff", invocation



import collections
import os.path
import hashlib
import json
from ruamel.yaml.comments import CommentedMap
from .tosca import ToscaSpec, TOSCA_VERSION

from .support import ResourceChanges, AttributeManager, Status, Priority, Action, Defaults
from .runtime import OperationalInstance, Resource, Capability, Relationship
from .util import UnfurlError, toEnum
from .configurator import Dependency, ConfigurationSpec
from .repo import RevisionManager, findGitRepo
from .yamlloader import YamlConfig, loadFromRepo
from .job import ConfigChange

import logging
logger = logging.getLogger('unfurl')

ChangeRecordAttributes = CommentedMap([
   ('changeId', 0),
   ('parentId', None),
   ('commitId', ''),
   ('startTime', ''),
]);

class Manifest(AttributeManager):
  """
  Loads a model from dictionary representing the manifest
  """
  def __init__(self, spec, path, localEnv=None):
    super(Manifest, self).__init__()
    self.localEnv = localEnv
    self.repo = localEnv and localEnv.instanceRepo
    self.currentCommitId = self.repo and self.repo.revision
    self.tosca = self.loadSpec(spec, path)
    self.specDigest = self.getSpecDigest(spec)
    self.revisions = RevisionManager(self)
    self.path = path

  def loadSpec(self, spec, path):
    if 'tosca' in spec:
      toscaDef = spec['tosca']
    elif 'node_templates' in spec:
      # allow node_templates shortcut
      toscaDef = {'node_templates': spec['node_templates']}
    else:
      toscaDef = {}
    if "node_templates" in toscaDef:
      # shortcut
      toscaDef = dict(tosca_definitions_version=TOSCA_VERSION,
                  topology_template=toscaDef)
    else:
      # make sure this is present
      toscaDef['tosca_definitions_version']=TOSCA_VERSION

    # hack so we can sneak through manifest to the yamlloader
    toscaDef = CommentedMap(toscaDef.items())
    toscaDef.manifest = self
    return ToscaSpec(toscaDef, spec.get('inputs'), spec, path)

  def getSpecDigest(self, spec):
    m = hashlib.sha1() # use same digest function as git
    t = self.tosca.template
    for tpl in [spec, t.topology_template.custom_defs, t.nested_tosca_tpls_with_topology]:
      m.update(json.dumps(tpl, sort_keys=True).encode("utf-8"))
    return m.hexdigest()

  def _ready(self, rootResource, lastChangeId=0):
    self.rootResource = rootResource
    if rootResource:
      rootResource.attributeManager = self
    self.lastChangeId = lastChangeId

  def getRootResource(self):
    return self.rootResource

  def getBaseDir(self):
    return '.'

  def saveJob(self, job):
    pass

  def loadTemplate(self, name, lastChange=None):
    if lastChange:
      return self.revisions.getRevision(lastChange).tosca.getTemplate(name)
    else:
      return self.tosca.getTemplate(name)

#  load instances
#    create a resource with the given template
#  or generate a template setting interface with the referenced implementations

  @staticmethod
  def loadStatus(status, instance=None):
    if not instance:
      instance = OperationalInstance()
    if not status:
      return instance

    instance._priority = toEnum(Priority, status.get('priority'))
    instance._lastStateChange = status.get('lastStateChange')
    instance._lastConfigChange = status.get('lastConfigChange')

    readyState = status.get('readyState')
    if not isinstance(readyState, collections.Mapping):
      instance._localStatus = toEnum(Status, readyState)
    else:
      instance._localStatus = toEnum(Status, readyState.get('local'))

    return instance

  @staticmethod
  def loadResourceChanges(changes):
    resourceChanges = ResourceChanges()
    if changes:
      for k, change in changes.items():
        status = change.pop('.status', None)
        resourceChanges[k] = [
          None if status is None else Manifest.loadStatus(status).localStatus,
          change.pop('.added', None),
          change
        ]
    return resourceChanges

  def loadConfigChange(self, changeId):
    """
    Reconstruct the Configuration that was applied in the past
    """
    changeSet = self.changeSets.get(changeId)
    if not changeSet:
      raise UnfurlError("can not find changeset for changeid %s" % changeId)

    configChange = ConfigChange()
    Manifest.loadStatus(changeSet, configChange)
    for (k,v) in ChangeRecordAttributes.items():
      setattr(self, k, changeSet.get(k, v))

    configChange.inputs = changeSet.get('inputs')

    configChange.dependencies = {}
    for val in changeSet.get('dependencies', []):
      key = val.get('name') or val['ref']
      assert key not in configChange.dependencies
      configChange.dependencies[key] = Dependency(val['ref'], val.get('expected'),
        val.get('schema'), val.get('name'), val.get('required'), val.get('wantList', False))

    if 'changes' in changeSet:
      configChange.resourceChanges = self.loadResourceChanges(changeSet['changes'])

    configChange.result = changeSet.get('result')
    configChange.messages = changeSet.get('messages', [])

    # XXX
    # ('action', ''),
    # ('target', ''), # nodeinstance key
    # implementationType: configurator resource | artifact | configurator class
    # implementation: repo:key#commitid | className:version
    return configChange

  # find config spec from potentially old version of the tosca template
  # get template then get node template name
  # but we shouldn't need this, except maybe to revert?
  def loadConfigSpec(self, configName, spec):
    return ConfigurationSpec(configName, spec['action'], spec['className'],
          spec.get('majorVersion'), spec.get('minorVersion',''),
          intent=toEnum(Action, spec.get('intent', Defaults.intent)),
          inputs=spec.get('inputs'), inputSchema=spec.get('inputSchema'),
          preConditions=spec.get('preConditions'), postConditions=spec.get('postConditions'))

  def loadResource(self, rname, resourceSpec, parent=None):
    # if parent property is set it overrides the parent argument
    pname = resourceSpec.get('parent')
    if pname:
      parent = self.getRootResource().findResource(pname)
      if parent is None:
        raise UnfurlError('can not find parent resource %s' % pname)

    resource = self._createNodeInstance(Resource, rname, resourceSpec, parent)
    return resource

  def _getLastChange(self, operational):
    if not operational.lastConfigChange:
      return None
    changerecord = self.changeSets[operational.lastConfigChange]
    parentId = changerecord.get('parentId')
    if parentId:
      return self.changeSets[parentId]
    return changerecord

  def _createNodeInstance(self, ctor, name, status, parent):
    operational = self.loadStatus(status)
    templateName = status.get('template', name)
    template = self.loadTemplate(templateName)
    if template is None:
      if operational.lastConfigChange:
        changerecord = self._getLastChange(operational)
        template = self.loadTemplate(templateName, changerecord)
    if template is None:
      raise UnfurlError('missing resource template %s' % templateName)
    logger.debug('template %s: %s', templateName, template)

    resource = ctor(name, status.get('attributes'), parent, template, operational)

    for key, val in status.get('capabilities', {}).items():
      self._createNodeInstance(Capability, key, val, resource)

    for key, val in status.get('requirements', {}).items():
      self._createNodeInstance(Relationship, key, val, resource)

    for key, val in status.get('resources', {}).items():
      self._createNodeInstance(Resource, key, val, resource)

    return resource

  def findRepoFromGitUrl(self, path, isFile=True, importLoader=None, willUse=False):
    repoURL, filePath, revision = findGitRepo(path, isFile, importLoader)
    if not repoURL or not self.localEnv:
      return None, None, None, None
    explicitRevision = revision
    basePath = importLoader.path #XXX check if dir or not
    # if not explicitRevision:
    #   revision = self.repoStatus.get(repoURL)
    repo, filePath, revision, bare = self.localEnv.findOrCreateWorkingDir(repoURL, isFile, revision, basePath)
    # if willUse and (not explicitRevision or not bare):
    #   self.updateRepoStatus(repo, revision)
    return repo, filePath, revision, bare

  def findPathInRepos(self, path, importLoader=None, willUse=False):
    """
    Check if the file path is inside a folder that is managed by a repository.
    If the revision is pinned and doesn't match the repo, it might be bare
    """
    candidate = None
    if self.repo: # our own repo gets first crack
      filePath, revision, bare = self.repo.findPath(path, importLoader)
      if filePath:
        if not bare:
          return self.repo, filePath, revision, bare
        else:
          candidate = (self.repo, filePath, revision, bare)
    if self.localEnv:
      repo, filePath, revision, bare = self.localEnv.findPathInRepos(path, importLoader)
      if repo:
        if bare and candidate:
          return candidate
        else:
          # if willUse:
          #     self.updateRepoStatus(repo, revision)
          return repo, filePath, revision, bare
    return None, None, None, None

  # def updateRepoStatus(self, repos):
  #   self.repoStatus.update({repo['url'] : repo['commit']
  #       for repo in repos.values() if repo.get('commit')})

  def loadHook(self, yamlConfig, templatePath, baseDir):
    #self.updateRepoStatus(yamlConfig.config.get('status', {}).get('repositories',{}))
    repositories = yamlConfig.config.get('spec', {}).get('tosca', {}).get('repositories', {})

    if isinstance(templatePath, dict):
      templatePath = templatePath.copy()
      path = templatePath['file']
      name = os.path.basename(path)
      repo = templatePath.get('repository')
      if isinstance(repo, dict):
        # a full repository spec maybe part of the include
        reponame = repo.pop('name', name)
        # replace spec with just its name
        templatePath['repository'] = reponame
        repositories[reponame] = repo
    else:
      name = os.path.basename(templatePath)
      templatePath = dict(file = templatePath)

    return loadFromRepo(name, templatePath, baseDir, repositories, self)

class SnapShotManifest(Manifest):
  def __init__(self, manifest, commitId):
    self.commitId = commitId
    oldManifest = manifest.repo.show(manifest.path, commitId)
    self.repo = manifest.repo
    self.localEnv =  manifest.localEnv
    self.manifest = YamlConfig(oldManifest, manifest.path,
                                    loadHook=self.loadHook)
    expanded = self.manifest.expanded
    spec = expanded.get('spec', {})
    super(SnapShotManifest, self).__init__(spec, self.manifest.path, self.localEnv)
    # just needs the spec, not root resource
    self._ready(None)



"""
inputs:
 command: "--switch {{ '.::foo' | ref }}"
 timeout: 9999
 resultTemplate:
  # cmd, stdout, stderr
  ref:
    file:
      ./handleResult.tpl
  select: contents
"""

# support tosca 4.2 Environment Variable Conventions (p 153)
# at least expose config parameters
# see also 13.3.1 Shell scripts p 328
# XXX add support for a stdin parameter
# (that's a reason to make config inputs lazy too)

from unfurl.configurator import Configurator, Status
import os
#import os.path
import sys
import six
if os.name == 'posix' and sys.version_info[0] < 3:
    import subprocess32 as subprocess
else:
    import subprocess
# cf https://github.com/opsmop/opsmop/blob/master/opsmop/core/command.py

import logging
logger = logging.getLogger('unfurl')

try:
  from shutil import which
except ImportError:
  from distutils import spawn
  def which(executable, mode=os.F_OK | os.X_OK, path=None):
    executable = spawn.find_executable(executable, path)
    if executable:
      if os.access(executable, mode):
        return executable
    return None

# XXX set environment vars
# XXX we should know if cmd if not os.access(implementation, os.X):
class ShellConfigurator(Configurator):

  def runProcess(self, cmd, shell=False, timeout=None):
    """
    Returns an object with the following attributes:

    cmd
    timeout (None unless timeout occurred)
    stderr
    stdout
    returncode (None if the process didn't complete)
    error if an exception was raised
    """
    if isinstance(cmd, list):
      cmdStr = " ".cmd.join(cmd)
    else:
      cmdStr = cmd

    try:
      completed = subprocess.run(cmd, shell=shell, timeout=timeout,
                      stdout=subprocess.PIPE, stderr=subprocess.PIPE)
      # try to convert stdout and stderr to strings but leave as binary if that fails
      try:
        completed.stdout = completed.stdout.decode()
      except:
        pass
      try:
        completed.stderr = completed.stderr.decode()
      except:
        pass
      completed.cmd = cmdStr
      completed.timeout = None
      completed.error = None
      return completed
    except subprocess.TimeoutExpired as err:
      completed.timeout = timeout
      err.returncode = None
      err.error = None
      return err
    except Exception as err:
      err.cmd = cmdStr
      err.timeout = None
      err.stderr = None
      err.stdout = None
      err.returncode = None
      err.error = err
      return err

  def handleResult(self, task, result, resultTemplate=None):
    status = Status.error if result.error or result.returncode else Status.ok
    if status == Status.error:
      logger.warning("shell task failed %s", result)
    else:
      logger.info("ran shell task %s", result)

    if status != Status.error and resultTemplate:
      results = task.query({
        'eval': dict(template=resultTemplate),
        'vars': result.__dict__})
      if results and results.strip():
        task.updateResources(results)
    return status

  def cantRun(self, task):
    params = task.inputs
    cmd = params.get('command')
    if not cmd:
      return "missing command to execute"
    if isinstance(cmd, list) and not params.get('shell') and not which(cmd[0]):
      return "'%s' is not executable" % cmd[0]
    return False

  def run(self, task):
    params = task.inputs
    assert not self.cantRun(task)
    cmd = params['command']
    # default for shell: True if command is a string otherwise False
    shell = params.get('shell', isinstance(cmd, six.string_types))
    result = self.runProcess(cmd, shell=shell, timeout=params.get('timeout'))
    status = self.handleResult(task, result, params.get('resultTemplate'))
    yield task.createResult(True, True, status, result=result.__dict__)



import unittest
from unfurl.runtime import Status, Priority, Resource, OperationalInstance
from unfurl.job import JobOptions, Runner
from unfurl.configurator import Configurator, ConfigurationSpec
from unfurl.yamlmanifest import YamlManifest
from unfurl.yamlloader import YamlConfig
from unfurl.util import UnfurlError, UnfurlValidationError, expandDoc, restoreIncludes, lookupClass, VERSION, diffDicts, mergeDicts, patchDict
from ruamel.yaml.comments import CommentedMap
import traceback
import six
import datetime
import copy
import os.path
from click.testing import CliRunner

class SimpleConfigurator(Configurator):
  def run(self, task):
    assert not self.cantRun(task)
    yield task.createResult(True, True, Status.ok)

simpleConfigSpec = ConfigurationSpec('subtask', 'instantiate', 'SimpleConfigurator', 0)

class TestSubtaskConfigurator(Configurator):

  def run(self, task):
    assert not self.cantRun(task)
    configuration = yield task.createSubTask(simpleConfigSpec)
    assert configuration.status == Status.ok, configuration.status
    # print ("running TestSubtaskConfigurator")
    yield task.createResult(True, True, Status.ok)

class ExpandDocTest(unittest.TestCase):
  doc =  {
    't1': {"b": 2},

    "t2": {
      'a': { 'b': 1},
      'c': 'c'
    },

    't3': 'val',

    't4': ['a', 'b'],

    'test1': CommentedMap([
     ('+t2', None),
     ('a', {'+t1': None }),
     ('d', {'+t3': None }),
     ('e', 'e')
    ]),

    'test2':
      [1, "+t1", '+t4', {'+t4': None}],
  }

  expected = {
    'test1': {
      'a': { 'b': 2},
      'c': 'c',
      'd': 'val',
      'e': 'e'
    },
    'test2':
      [1, {"b": 2}, 'a', 'b', 'a', 'b']
  }

  def test_expandDoc(self):
    includes, expanded = expandDoc(self.doc, cls=CommentedMap)
    self.assertEqual(includes, {
      ('test1',): [('+t2', None)],
      ('test1', 'a'): [('+t1', None)],
      ('test1', 'd'): [('+t3', None)],
      ('test2', 1): [('+t1', None)],
      ('test2', 2): [('+t4', None)],
      ('test2', 3): [('+t4', None)],
    })
    self.assertEqual(expanded['test1'], self.expected['test1'])
    self.assertEqual(expanded['test2'], self.expected['test2'])
    restoreIncludes(includes, self.doc, expanded, CommentedMap)
    self.assertEqual(expanded['test1'], self.doc['test1'])

  def test_diff(self):
    expectedOld = {'a': 1, 'b': {'b1': 1, 'b2': 1}, 'd':1}
    old = copy.copy(expectedOld)
    new = {'a': 1, 'b': {'b1': 2, 'b2': 1}, 'c': 2}
    diff = diffDicts(old, new)
    self.assertEqual(diff, {'b': {'b1': 2}, 'c': 2, 'd': {'+%': 'delete'}})
    newNew = mergeDicts(old, diff)
    self.assertEqual(newNew, new)
    self.assertEqual(old, expectedOld)
    patchDict(old, new)
    self.assertEqual(old, new)

  def test_missingInclude(self):
    doc1 = CommentedMap([
     ('+a/c', None),
     ('a', {'+b': None }),
     ('b', {'c': {'d': 1}}),
    ])
    includes, expanded = expandDoc(doc1, cls=CommentedMap)
    self.assertEqual(expanded, {
      'd': 1,
      'a': {'c': {'d': 1}},
      'b': {'c': {'d': 1}}
    })
    self.assertEqual(len(includes), 2)

    doc1['missing'] = {'+include-missing': None}
    with self.assertRaises(UnfurlError) as err:
      includes, expanded = expandDoc(doc1, cls=CommentedMap)

  def test_recursion(self):
    doc = {
      "test3": {
        "a": {
          "recurse": {"+test3": None}
        }
      },
    }
    with self.assertRaises(UnfurlError) as err:
      includes, expanded = expandDoc(doc)
    self.assertEqual(str(err.exception),
      '''recursive include "['test3']" in "('test3', 'a', 'recurse')"''')

    # XXX fix ../
    # doc2 = {
    #   "test4": {
    #     "+../test4": None,
    #     "child": {}
    #   }
    # }
    # includes, expanded = expandDoc(doc2)

class JobTest(unittest.TestCase):

  def test_lookupClass(self):
    assert lookupClass("SimpleConfigurator") is SimpleConfigurator

  # XXX rewrite
  # def test_runner(self):
  #   rootResource = Resource('root')
  #   configurationSpec = ConfigurationSpec('test', 'instantiate', 'TestSubtaskConfigurator', 0)
  #   specs = [configurationSpec]
  #   runner = Runner(Manifest(rootResource, specs))
  #   job = runner.run(JobOptions(add=True))
  #   assert not job.unexpectedAbort, job.unexpectedAbort.stackInfo

class OperationalInstanceTest(unittest.TestCase):

  def test_aggregate(self):
    ignoredError = OperationalInstance("error", "ignore")
    assert ignoredError.status == Status.error
    assert ignoredError.priority == Priority.ignore
    assert not ignoredError.operational
    assert not ignoredError.required

    requiredError = OperationalInstance("error", "required")
    assert requiredError.status == Status.error
    assert requiredError.priority == Priority.required
    assert requiredError.required
    assert not requiredError.operational

    aggregateOK = OperationalInstance(Status.ok)
    aggregateOK.dependencies = [ignoredError]
    assert aggregateOK.status == Status.ok

    aggregateError = OperationalInstance(Status.ok)
    aggregateError.dependencies = [ignoredError, requiredError]
    self.assertEqual(aggregateError.status, Status.error)

    aggregateError = OperationalInstance(Status.notapplied)
    aggregateError.dependencies = [ignoredError, requiredError]
    self.assertEqual(aggregateError.status, Status.error)

    aggregateError = OperationalInstance(Status.notapplied)
    aggregateError.dependencies = [OperationalInstance("ok", "optional")]
    self.assertEqual(aggregateError.status, Status.ok)

    aggregateError = OperationalInstance(Status.notapplied)
    aggregateError.dependencies = []
    self.assertEqual(aggregateError.status, Status.notapplied)

class RunTest(unittest.TestCase):
  """
  XXX:
  test status: report last state, report config changes, report plan
  test read-ony/discover
  test failing configurations with abort, continue, revert (both capable of reverting and not)
  test incomplete runs
  test unexpected errors, terminations
  test locking
  test required metadata on resources
  """
  # XXX3 test case: a failed configuration with intent=revert should stay in configurations as error
  # XXX3 test: hide configurations that are both notpresent and revert / skip

  def test_manifest(self):
    simple = """
    apiVersion: %s
    kind: Manifest
    spec:
      instances:
        anInstance:
          # template: foo
          # or use shortcut
          implementations:
            create: test
      implementations:
        test:
          className:    TestSubtaskConfigurator
          inputs: {}
    """ % VERSION
    manifest = YamlManifest(simple)
    runner = Runner(manifest)
    self.assertEqual(runner.lastChangeId, 0)
    output = six.StringIO()
    job = runner.run(JobOptions(add=True, out=output, startTime="test"))
    assert not job.unexpectedAbort, job.unexpectedAbort.getStackTrace()
    # workDone includes subtasks
    assert len(job.workDone) == 2, job.workDone

    # manifest shouldn't have changed
    manifest2 = YamlManifest(output.getvalue())
    self.assertEqual(manifest2.lastChangeId, 3)
    output2 = six.StringIO()
    job2 = Runner(manifest2).run(JobOptions(add=True, out=output2, startTime="test"))
    # print('2', output2.getvalue())
    assert not job2.unexpectedAbort, job2.unexpectedAbort.getStackTrace()

    # should not find any tasks to run
    assert len(job2.workDone) == 0, job2.workDone
    self.maxDiff = None
    self.assertEqual(output.getvalue(), output2.getvalue())

  def test_template_inheritance(self):
    manifest = '''
apiVersion: unfurl/v1alpha1
kind: Manifest
configurators:
  step1:
    spec:
      className: "TestSubtaskConfigurator"
      majorVersion: 0
templates:
  base:
    configurations:
      step1:
        +configurators/step1:
root:
  resources:
    cloud3: #key is resource name
      +templates/base:
      +templates/production:
'''
    with self.assertRaises(UnfurlError) as err:
      YamlManifest(manifest)
    self.assertIn('missing includes: [templates/production]', str(err.exception))

class TestInterface:
  def __init__(self, interfaceName, resource):
    self.name = interfaceName
    self.resource = resource

class InterfaceTest(unittest.TestCase):
  """
  .interfaces:
    inherit: ClassName
    default:
    filter:
  """

  def test_interface(self):
    r = Resource('test')
    r.addInterface(TestInterface)
    className = __name__ + '.TestInterface'
    self.assertEqual(r.attributes['.interfaces'], {className: className})
    i = r.getInterface(className)
    assert i, "interface not found"
    self.assertIs(r, i.resource)
    self.assertEqual(i.name, className)
    self.assertIs(r._interfaces[className], i)

  # XXX test ::container[.interfaces=Container]

class FileTestConfigurator(Configurator):
  def run(self, task):
    assert not self.cantRun(task)
    assert task.target.attributes['file'] == 'foo.txt', task.target.attributes
    assert os.path.isabs(task.inputs['path']), task.inputs
    assert task.inputs['contents'] == 'test', task.inputs['contents']

    filevalue = task.query({'ref':{'file':'foo.txt'}}, wantList=True)
    assert filevalue._attributes[0].external.type == 'file'

    filevalue = task.query({'ref':{'file':'foo.txt'}, 'foreach': 'path'}, wantList=True)
    assert filevalue._attributes[0].external.type == 'file'

    value = task.query({'ref':{'file':'foo.txt'}})
    yield task.createResult(True, True, Status.ok, result = value)

class FileTest(unittest.TestCase):

  def test_fileRef(self):
    simple = """
    apiVersion: %s
    kind: Manifest
    spec:
      node_templates:
        test:
          type: tosca.nodes.Root
          properties:
            file:
              eval:
                file: foo.txt
          interfaces:
            Standard:
                create: test
      implementations:
        test:
          className:    FileTestConfigurator
          inputs:
            path:
              ref: file::path
            contents:
              ref: file::contents
    """ % VERSION
    manifest = YamlManifest(simple)
    runner = Runner(manifest)
    output = six.StringIO()
    cliRunner = CliRunner()
    with cliRunner.isolated_filesystem(): # as tmpDir
      with open('foo.txt', 'w') as f:
        f.write('test')
      job = runner.run(JobOptions(add=True, out=output, startTime="test"))
    task = list(job.workDone.values())[0]
    self.assertEqual(task.result.result, 'foo.txt')

  def test_manifest_template(self):
    template = """
apiVersion: unfurl/v1alpha1
kind: Manifest
spec:
  a: 1
  b: 2
status: {}
    """
    cliRunner = CliRunner()
    with cliRunner.isolated_filesystem(): # as tmpDir
      with open('template.yaml', 'w') as f:
        f.write(template)

      instanceYaml = """
apiVersion: unfurl/v1alpha1
kind: Manifest
+%include:
  file: template.yaml
spec:
  b: 3
      """
      manifest = YamlManifest(instanceYaml)
      assert manifest.manifest.expanded['spec']['a'] == 1
      assert manifest.manifest.expanded['spec']['b'] == 3
      output = six.StringIO()
      manifest.dump(output)
      config = YamlConfig(output.getvalue())
      assert config.config['+%include'] == {'file':'template.yaml'}
      assert 'a' not in config.config['spec']

  def test_change(self):
    """
    config parameter: file.path
    run...
    touch file...
    run again...
    assert it triggers update
    """
    # XXX

class ImportTestConfigurator(Configurator):
  def run(self, task):
    assert not self.cantRun(task)
    assert task.target.attributes['test']
    assert task.target.attributes['mapped1']
    yield task.createResult(True, True, Status.ok)

class ImportTest(unittest.TestCase):

  def test_import(self):
    foreign = """
    apiVersion: %s
    kind: Manifest
    spec:
     instances:
      foreign:
        attributes:
          prop1: ok
          prop2: not-a-number
    """ % VERSION
    runner = CliRunner()
    with runner.isolated_filesystem():
      with open('foreignmanifest.yaml', 'w') as f:
        f.write(foreign)

      importer = """
apiVersion: %s
kind: Manifest
imports:
  test:
    file: foreignmanifest.yaml
    resource: foreign # default is root
    # attributes: # queries into resource
    properties: # expected schema for attributes
      prop2:
       type: number
      prop3:
        type: string
        default: 'default'
spec:
  instances:
    importer:
  node_templates:
    importer:
      type: tosca.nodes.Root
      properties:
            test:
              eval:
                external: test
            mapped1:
              eval:
                external: test
              foreach: prop1
            mapped2:
              eval:
                external: test
              foreach: prop2
            mapped3:
              eval:
                external: test
              foreach: prop3
      interfaces:
        Standard:
            create: ImportTestConfigurator
      """ % VERSION
      manifest = YamlManifest(importer)
      root = manifest.getRootResource()
      importerResource = root.findResource('importer')
      #assert importing.attributes['test']
      assert root.imports['test']
      self.assertEqual(importerResource.attributes['mapped1'], 'ok')
      with self.assertRaises(UnfurlValidationError) as err:
        importerResource.attributes['mapped2']
      self.assertIn('schema validation failed', str(err.exception))
      self.assertEqual(importerResource.attributes['mapped3'], 'default')
      job = Runner(manifest).run(JobOptions(add=True, startTime="time-to-test"))
      assert job.status == Status.ok, job.summary()



import unittest
from unfurl.yamlmanifest import YamlManifest
from unfurl.job import Runner, JobOptions
from unfurl.configurator import Configurator, Status
from unfurl.util import lookupPath
import datetime

manifest = '''
apiVersion: unfurl/v1alpha1
kind: Manifest
configurations:
  create:
    implementation: unfurl.shellconfigurator.ShellConfigurator
    inputs:
     command: "echo 'helloworld'"
     timeout: 9999
     resultTemplate:
        q: |
           - name: .self
             attributes:
               stdout: "{{ stdout | trim }}"
spec:
  node_templates:
    test1:
      type: tosca.nodes.Root
      interfaces:
        Standard:
          +configurations:
'''

class ShellConfiguratorTest(unittest.TestCase):

  def test_shell(self):
    """
    test that runner figures out the proper tasks to run
    """
    runner = Runner(YamlManifest(manifest))

    run1 = runner.run(JobOptions(resource='test1'))
    assert len(run1.workDone) == 1, run1.workDone
    self.assertEqual(runner.manifest.getRootResource().findResource('test1').attributes['stdout'], 'helloworld')
    assert not run1.unexpectedAbort, run1.unexpectedAbort.getStackTrace()



# Copyright 2019, David Wilson
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# 1. Redistributions of source code must retain the above copyright notice,
# this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright notice,
# this list of conditions and the following disclaimer in the documentation
# and/or other materials provided with the distribution.
#
# 3. Neither the name of the copyright holder nor the names of its contributors
# may be used to endorse or promote products derived from this software without
# specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
# ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE
# LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
# CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
# SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
# INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
# CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
# ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
# POSSIBILITY OF SUCH DAMAGE.

from __future__ import absolute_import
import atexit
import errno
import logging
import os
import signal
import socket
import sys
import time

try:
    import faulthandler
except ImportError:
    faulthandler = None

import mitogen
import mitogen.core
import mitogen.debug
import mitogen.master
import mitogen.parent
import mitogen.service
import mitogen.unix
import mitogen.utils

import ansible
import ansible.constants as C
import ansible_mitogen.logging
import ansible_mitogen.services

from mitogen.core import b
import ansible_mitogen.affinity


LOG = logging.getLogger(__name__)

ANSIBLE_PKG_OVERRIDE = (
    u"__version__ = %r\n"
    u"__author__ = %r\n"
)


def clean_shutdown(sock):
    """
    Shut the write end of `sock`, causing `recv` in the worker process to wake
    up with a 0-byte read and initiate mux process exit, then wait for a 0-byte
    read from the read end, which will occur after the the child closes the
    descriptor on exit.

    This is done using :mod:`atexit` since Ansible lacks any more sensible hook
    to run code during exit, and unless some synchronization exists with
    MuxProcess, debug logs may appear on the user's terminal *after* the prompt
    has been printed.
    """
    sock.shutdown(socket.SHUT_WR)
    sock.recv(1)


def getenv_int(key, default=0):
    """
    Get an integer-valued environment variable `key`, if it exists and parses
    as an integer, otherwise return `default`.
    """
    try:
        return int(os.environ.get(key, str(default)))
    except ValueError:
        return default


def save_pid(name):
    """
    When debugging and profiling, it is very annoying to poke through the
    process list to discover the currently running Ansible and MuxProcess IDs,
    especially when trying to catch an issue during early startup. So here, if
    a magic environment variable set, stash them in hidden files in the CWD::

        alias muxpid="cat .ansible-mux.pid"
        alias anspid="cat .ansible-controller.pid"

        gdb -p $(muxpid)
        perf top -p $(anspid)
    """
    if os.environ.get('MITOGEN_SAVE_PIDS'):
        with open('.ansible-%s.pid' % (name,), 'w') as fp:
            fp.write(str(os.getpid()))


class MuxProcess(object):
    """
    Implement a subprocess forked from the Ansible top-level, as a safe place
    to contain the Mitogen IO multiplexer thread, keeping its use of the
    logging package (and the logging package's heavy use of locks) far away
    from the clutches of os.fork(), which is used continuously by the
    multiprocessing package in the top-level process.

    The problem with running the multiplexer in that process is that should the
    multiplexer thread be in the process of emitting a log entry (and holding
    its lock) at the point of fork, in the child, the first attempt to log any
    log entry using the same handler will deadlock the child, as in the memory
    image the child received, the lock will always be marked held.

    See https://bugs.python.org/issue6721 for a thorough description of the
    class of problems this worker is intended to avoid.
    """

    #: In the top-level process, this references one end of a socketpair(),
    #: which the MuxProcess blocks reading from in order to determine when
    #: the master process dies. Once the read returns, the MuxProcess will
    #: begin shutting itself down.
    worker_sock = None

    #: In the worker process, this references the other end of
    #: :py:attr:`worker_sock`.
    child_sock = None

    #: In the top-level process, this is the PID of the single MuxProcess
    #: that was spawned.
    worker_pid = None

    #: A copy of :data:`os.environ` at the time the multiplexer process was
    #: started. It's used by mitogen_local.py to find changes made to the
    #: top-level environment (e.g. vars plugins -- issue #297) that must be
    #: applied to locally executed commands and modules.
    original_env = None

    #: In both processes, this is the temporary UNIX socket used for
    #: forked WorkerProcesses to contact the MuxProcess
    unix_listener_path = None

    #: Singleton.
    _instance = None

    @classmethod
    def start(cls, _init_logging=True):
        """
        Arrange for the subprocess to be started, if it is not already running.

        The parent process picks a UNIX socket path the child will use prior to
        fork, creates a socketpair used essentially as a semaphore, then blocks
        waiting for the child to indicate the UNIX socket is ready for use.

        :param bool _init_logging:
            For testing, if :data:`False`, don't initialize logging.
        """
        if cls.worker_sock is not None:
            return

        if faulthandler is not None:
            faulthandler.enable()

        mitogen.utils.setup_gil()
        cls.unix_listener_path = mitogen.unix.make_socket_path()
        cls.worker_sock, cls.child_sock = socket.socketpair()
        atexit.register(lambda: clean_shutdown(cls.worker_sock))
        mitogen.core.set_cloexec(cls.worker_sock.fileno())
        mitogen.core.set_cloexec(cls.child_sock.fileno())

        cls.profiling = os.environ.get('MITOGEN_PROFILING') is not None
        if cls.profiling:
            mitogen.core.enable_profiling()
        if _init_logging:
            ansible_mitogen.logging.setup()

        cls.original_env = dict(os.environ)
        cls.child_pid = os.fork()
        if cls.child_pid:
            save_pid('controller')
            ansible_mitogen.logging.set_process_name('top')
            ansible_mitogen.affinity.policy.assign_controller()
            cls.child_sock.close()
            cls.child_sock = None
            mitogen.core.io_op(cls.worker_sock.recv, 1)
        else:
            save_pid('mux')
            ansible_mitogen.logging.set_process_name('mux')
            ansible_mitogen.affinity.policy.assign_muxprocess()
            cls.worker_sock.close()
            cls.worker_sock = None
            self = cls()
            self.worker_main()

    def worker_main(self):
        """
        The main function of for the mux process: setup the Mitogen broker
        thread and ansible_mitogen services, then sleep waiting for the socket
        connected to the parent to be closed (indicating the parent has died).
        """
        self._setup_master()
        self._setup_services()

        try:
            # Let the parent know our listening socket is ready.
            mitogen.core.io_op(self.child_sock.send, b('1'))
            # Block until the socket is closed, which happens on parent exit.
            mitogen.core.io_op(self.child_sock.recv, 1)
        finally:
            self.broker.shutdown()
            self.broker.join()

            # Test frameworks living somewhere higher on the stack of the
            # original parent process may try to catch sys.exit(), so do a C
            # level exit instead.
            os._exit(0)

    def _enable_router_debug(self):
        if 'MITOGEN_ROUTER_DEBUG' in os.environ:
            self.router.enable_debug()

    def _enable_stack_dumps(self):
        secs = getenv_int('MITOGEN_DUMP_THREAD_STACKS', default=0)
        if secs:
            mitogen.debug.dump_to_logger(secs=secs)

    def _setup_simplejson(self, responder):
        """
        We support serving simplejson for Python 2.4 targets on Ansible 2.3, at
        least so the package's own CI Docker scripts can run without external
        help, however newer versions of simplejson no longer support Python
        2.4. Therefore override any installed/loaded version with a
        2.4-compatible version we ship in the compat/ directory.
        """
        responder.whitelist_prefix('simplejson')

        # issue #536: must be at end of sys.path, in case existing newer
        # version is already loaded.
        compat_path = os.path.join(os.path.dirname(__file__), 'compat')
        sys.path.append(compat_path)

        for fullname, is_pkg, suffix in (
            (u'simplejson', True, '__init__.py'),
            (u'simplejson.decoder', False, 'decoder.py'),
            (u'simplejson.encoder', False, 'encoder.py'),
            (u'simplejson.scanner', False, 'scanner.py'),
        ):
            path = os.path.join(compat_path, 'simplejson', suffix)
            fp = open(path, 'rb')
            try:
                source = fp.read()
            finally:
                fp.close()

            responder.add_source_override(
                fullname=fullname,
                path=path,
                source=source,
                is_pkg=is_pkg,
            )

    def _setup_responder(self, responder):
        """
        Configure :class:`mitogen.master.ModuleResponder` to only permit
        certain packages, and to generate custom responses for certain modules.
        """
        responder.whitelist_prefix('ansible')
        responder.whitelist_prefix('ansible_mitogen')
        self._setup_simplejson(responder)

        # Ansible 2.3 is compatible with Python 2.4 targets, however
        # ansible/__init__.py is not. Instead, executor/module_common.py writes
        # out a 2.4-compatible namespace package for unknown reasons. So we
        # copy it here.
        responder.add_source_override(
            fullname='ansible',
            path=ansible.__file__,
            source=(ANSIBLE_PKG_OVERRIDE % (
                ansible.__version__,
                ansible.__author__,
            )).encode(),
            is_pkg=True,
        )

    def _setup_master(self):
        """
        Construct a Router, Broker, and mitogen.unix listener
        """
        self.broker = mitogen.master.Broker(install_watcher=False)
        self.router = mitogen.master.Router(
            broker=self.broker,
            max_message_size=4096 * 1048576,
        )
        self._setup_responder(self.router.responder)
        mitogen.core.listen(self.broker, 'shutdown', self.on_broker_shutdown)
        mitogen.core.listen(self.broker, 'exit', self.on_broker_exit)
        self.listener = mitogen.unix.Listener(
            router=self.router,
            path=self.unix_listener_path,
            backlog=C.DEFAULT_FORKS,
        )
        self._enable_router_debug()
        self._enable_stack_dumps()

    def _setup_services(self):
        """
        Construct a ContextService and a thread to service requests for it
        arriving from worker processes.
        """
        self.pool = mitogen.service.Pool(
            router=self.router,
            services=[
                mitogen.service.FileService(router=self.router),
                mitogen.service.PushFileService(router=self.router),
                ansible_mitogen.services.ContextService(self.router),
                ansible_mitogen.services.ModuleDepService(self.router),
            ],
            size=getenv_int('MITOGEN_POOL_SIZE', default=32),
        )
        LOG.debug('Service pool configured: size=%d', self.pool.size)

    def on_broker_shutdown(self):
        """
        Respond to broker shutdown by beginning service pool shutdown. Do not
        join on the pool yet, since that would block the broker thread which
        then cannot clean up pending handlers, which is required for the
        threads to exit gracefully.
        """
        # In normal operation we presently kill the process because there is
        # not yet any way to cancel connect().
        self.pool.stop(join=self.profiling)

    def on_broker_exit(self):
        """
        Respond to the broker thread about to exit by sending SIGTERM to
        ourself. In future this should gracefully join the pool, but TERM is
        fine for now.
        """
        if not self.profiling:
            # In normal operation we presently kill the process because there is
            # not yet any way to cancel connect(). When profiling, threads
            # including the broker must shut down gracefully, otherwise pstats
            # won't be written.
            os.kill(os.getpid(), signal.SIGTERM)



import unittest
from unfurl.yamlmanifest import YamlManifest
from unfurl.job import Runner, JobOptions, Status
from unfurl.configurator import Configurator
from unfurl.util import lookupPath
import datetime

class TestConfigurator(Configurator):
  def run(self, task):
    assert not self.cantRun(task)
    attributes = task.target.attributes
    attributes['copyOfMeetsTheRequirement'] = attributes["meetsTheRequirement"]
    params = task.inputs
    if params.get('addresources'):
      shouldYield = params.get('yieldresources')
      resourceSpec = {
        'name': params['resourceName'],
        'template': 'test1'
      }
      if shouldYield:
        resourceSpec['dependent'] = True

      updateSpec = dict(name='.self', status=dict(attributes={'newAttribute': 1}))
      jobrequest = task.addResources([resourceSpec, updateSpec])
      if shouldYield:
        job = yield jobrequest
        assert job, jobrequest

    yield task.createResult(True, True, Status.ok)

manifest = '''
apiVersion: unfurl/v1alpha1
kind: Manifest
spec:
 implementations:
  test:
    apiVersion: unfurl/v1alpha1
    className: TestConfigurator
    majorVersion: 0
    preConditions:
      properties:
        meetsTheRequirement:
          type: string
      required: ['meetsTheRequirement']
 instances:
    test1:
        attributes:
          meetsTheRequirement: "copy"
        implementations:
          create: test
    test2:
        attributes:
          meetsTheRequirement: false
        implementations:
          create: test
    test3:
        +spec/instances/test1:
        implementations:
          create:
            implementation:
              primary: test
            inputs:
              resourceName: added1
              addresources: true
    test4:
        +spec/instances/test3:
        implementations:
          create:
            implementation:
              primary: test
            inputs:
              resourceName: added2
              yieldresources: true
'''

class ConfiguratorTest(unittest.TestCase):
  def verifyRoundtrip(self, original, jobOptions):
    job = Runner(YamlManifest(original)).run(jobOptions)
    # should not need to run any tasks
    assert len(job.workDone) == 0, job.workDone

    self.maxDiff = None
    self.assertEqual(original, job.out.getvalue())

  def test_neededTasks(self):
    """
    test that runner figures out the proper tasks to run
    """
    runner = Runner(YamlManifest(manifest))
    test1 = runner.manifest.getRootResource().findResource('test1')
    assert test1
    # missing = runner.manifest.spec[0].findInvalidPreconditions(test1)
    # assert not missing, missing

    run1 = runner.run(JobOptions(resource='test1'))
    assert len(run1.workDone) == 1, run1.workDone
    assert not run1.unexpectedAbort, run1.unexpectedAbort.getStackTrace()

  def test_preConditions(self):
    """test that the configuration only runs if the resource meets the requirements"""
    runner = Runner(YamlManifest(manifest))
    test1 = runner.manifest.getRootResource().findResource('test1')
    assert test1

    self.assertEqual(test1.attributes["meetsTheRequirement"], "copy")
#     notYetProvided = runner.manifest.spec[0].findMissingProvided(test1)
#     self.assertEqual(str(notYetProvided[1]),
# """[<ValidationError: "'copyOfMeetsTheRequirement' is a required property">]""")

    jobOptions1 = JobOptions(resource='test1', startTime=datetime.datetime.fromordinal(1))
    run1 = runner.run(jobOptions1)
    # print(run1.out.getvalue())
    assert not run1.unexpectedAbort, run1.unexpectedAbort.getStackTrace()
    self.assertEqual(test1.attributes['copyOfMeetsTheRequirement'], "copy")

    # provided = runner.manifest.spec[0].findMissingProvided(test1)
    # assert not provided, provided

    # check that the modifications were recorded
    self.assertEqual(runner.manifest.manifest.config['changes'][0]['changes'],
          {'::test1':{'copyOfMeetsTheRequirement': 'copy', '.status': 'ok'}})

    test2 = runner.manifest.getRootResource().findResource('test2')
    assert test2
    requiredAttribute = test2.attributes['meetsTheRequirement']
    assert requiredAttribute is False, requiredAttribute
    # missing = runner.manifest.specs[0].findMissingRequirements(test2)
    # self.assertEqual(str(missing[1]), '''[<ValidationError: "False is not of type 'string'">]''')

    self.verifyRoundtrip(run1.out.getvalue(), jobOptions1)

    jobOptions2 = JobOptions(resource='test2', startTime=datetime.datetime.fromordinal(1))
    run2 = runner.run(jobOptions2)
    assert not run2.unexpectedAbort, run2.unexpectedAbort.getStackTrace()
    assert run2.status == Status.error, run2.status
    # XXX better error reporting
    # self.assertEqual(str(run2.problems), "can't run required configuration: resource test2 doesn't meet requirement")

    # don't re-run the failed configurations so nothing will have changed
    jobOptions2.repair="none"
    self.verifyRoundtrip(run2.out.getvalue(), jobOptions2)

  @unittest.skip("fix adding resources api")
  def test_addingResources(self):
    runner = Runner(YamlManifest(manifest))
    jobOptions = JobOptions(resource='test3', startTime=datetime.datetime.fromordinal(1))
    run = runner.run(jobOptions)
    assert not run.unexpectedAbort, run.unexpectedAbort.getStackTrace()
    # self.assertEqual(list(run.workDone.keys()), [('test3', 'test'), ('added1', 'config1')])

    # verify added1
    added = {'.added': {
      'name': 'added1',
      'template': 'test1'
      }
    }
    modifications = lookupPath(runner.manifest.manifest.config,
    'changes.test.modifications'.split('.'))
    self.assertEqual(modifications['added1'], added)
    self.assertEqual(runner.manifest.manifest.config['changes'][0]['changes']['added1'], added)

    # verify modified
    self.assertEqual(modifications['test3'], {
      'copyOfMeetsTheRequirement': 'copy',
      'newAttribute': 1
    })

    # print('test3', run.out.getvalue())
    jobOptions.repair="none"
    self.verifyRoundtrip(run.out.getvalue(), jobOptions)

    jobOptions = JobOptions(resource='test4', startTime=datetime.datetime.fromordinal(1))
    run = runner.run(jobOptions)
    assert not run.unexpectedAbort, run.unexpectedAbort.getStackTrace()
    self.assertEqual(list(run.workDone.keys()), [('test4', 'test'), ('added2', 'config1')])
    # print('test4', run.out.getvalue())

    # verify dependencies added
    dependencies = lookupPath(runner.manifest.manifest.config,
    'root.resources.test4.status.configurations.test.dependencies'.split('.'))
    self.assertEqual(dependencies, [{"ref": "::added2"}])

    jobOptions.repair="none"
    self.verifyRoundtrip(run.out.getvalue(), jobOptions)

  # def test_shouldRun(self):
  #   pass
  #   #assert should_run
  #   # run()
  #   #assert not should_run
  #
  # def test_provides(self):
  #   #test that it provides as expected
  #   #test that there's an error if provides fails
  #   pass
  #
  # def test_update(self):
  #   #test version changed
  #   pass
  #
  # def test_configChanged(self):
  #   #test version changed
  #   pass
  #
  # def test_revert(self):
  #   # assert provides is removed
  #   pass



import collections
import json
import tempfile
import sys
from ruamel.yaml import YAML
from unfurl.util import ansibleDisplay, ansibleDummyCli, assertForm
from unfurl.configurator import Configurator, Status
import ansible.constants as C
from ansible.cli.playbook import PlaybookCLI
from ansible.plugins.callback.default import CallbackModule
from ansible.module_utils import six

import logging
logger = logging.getLogger('unfurl')

yaml = YAML()

# input parameters:
#  playbook
#  playbookArgs
#  extraVars
#  inventory
#  returnValues
#  XXX environVars
def ansibleResults(result, extraKeys=()):
  # result is per-task ansible.executor.task_result.TaskResult
  # https://github.com/ansible/ansible/blob/devel/lib/ansible/executor/task_result.py
  # https://docs.ansible.com/ansible/latest/reference_appendices/common_return_values.html
  # map to same result names used by shellconfigurator

  # XXX map ansible facts and user set variables, needed for result templates
  # https://docs.ansible.com/ansible/latest/user_guide/playbooks_variables.html#variables-discovered-from-systems-facts
  """
  stderr
  stdout
  returncode (None if the process didn't complete)
  error if an exception was raised
  """
  # _check_key checks 'results' if task was a loop
  # 'warnings': result._check_key('warning'),
  result = result.clean_copy()
  # print('result._result', result._result)
  resultDict = {
    'returncode': result._check_key('rc'),
  }
  keyMap = {'msg':['msg'], 'error': ['exception', 'module_stderr'], 'stdout': ['stdout', 'module_stdout']}
  for name, keys in keyMap.items():
    for key in keys:
      val = result._check_key(key)
      if val:
        resultDict[name] = val
        break

  for key in extraKeys:
    resultDict[key] = result._check_key(key)
  return resultDict

class AnsibleConfigurator(Configurator):
  """
  The current resource is the inventory.
  #could have parameter for mapping resource attributes to groups
  #also need to map attributes to host vars
  sshconfig
  ansible variables can not be set to a value of type resource

  external inventory discovers resources
  need away to map hosts to existing resources
  and to map vars to types of different resource
  """

  def __init__(self, configSpec):
    super(AnsibleConfigurator, self).__init__(configSpec)
    self._cleanupRoutines = []

  def dryRun(self, task):
    for result in self.run(task):
      yield result

  def getInventory(self, task):
    return task.inputs.get('inventory')
    # add groups for tosca HOST, ORCHESTRATOR, TARGET, SELF keywords
    # XXX cache and reuse file
    # https://github.com/ansible/ansible/issues/33132#issuecomment-346575458
    #return self._saveToTempfile(inventory, '-inventory.ini')
    # XXX if user set inventory create a folder to merge:
    # https://allandenot.com/devops/2015/01/16/ansible-with-multiple-inventory-files.html

  def _saveToTempfile(self, obj, suffix='.yml'):
    tp = tempfile.NamedTemporaryFile('w+t', suffix=suffix)
    self._cleanupRoutines.append(lambda: tp.close())
    if suffix.endswith('.json'):
      json.dump(obj, tp)
    elif suffix.endswith('.yml') or suffix.endswith('.yaml'):
      yaml.dump(obj, tp)
    else:
      tp.write(obj)
    return tp.name

  def _cleanup(self):
    for func in self._cleanupRoutines:
      try:
        func()
      except:
        # XXX: log
        pass
    self._cleanupRoutines = []

  # have a parameter for the mapping or just set_facts in playbook
  def getVars(self, task):
    """
    just add a lookup() plugin?
    add attribute to parameters to map to vars?
    # ansible yaml file but we evaluate the valuesFrom first?
    """
    return {}

  def _makePlayBook(self, playbook):
    assertForm(playbook, collections.MutableSequence)
    # XXX use host group instead of localhost depending on operation_host
    if playbook and not 'hosts' in playbook[0]:
      return [dict(hosts='localhost', gather_facts=False, tasks=playbook)]
    else:
      return playbook

  def findPlaybook(self, task):
    return task.inputs['playbook']

  def getPlaybook(self, task):
    playbook = self.findPlaybook(task)
    if isinstance(playbook, six.string_types):
      return playbook
    playbook = self._makePlayBook(playbook)
    return self._saveToTempfile(playbook, '-playbook.yml')

  def getPlaybookArgs(self, task):
    args = task.inputs.get('playbookArgs', [])
    if not isinstance(args, collections.MutableSequence):
      args = [args]
    if task.dryRun:
      args.append('--check')
    return args

  def _processResult(self, task, result):
    resultTemplate = task.inputs.get('resultTemplate')
    if resultTemplate:
      results = task.query({
        'eval': dict(template=resultTemplate),
        'vars': result.__dict__})
      if results and results.strip():
        task.updateResources(results)

  def getResultKeys(self, task, results):
    return task.inputs.get('returnValues', [])

  def run(self, task):
    try:
      #build host inventory from resource
      inventory = self.getInventory(task)
      playbook = self.getPlaybook(task)

      #build vars from inputs
      extraVars = self.getVars(task)
      results = runPlaybooks([playbook], inventory, extraVars, self.getPlaybookArgs(task))

      if results.exit_code or len(results.resultsByStatus.failed):
        status = Status.error
      else:
        # unreachable, failed, skipped
        # XXX degraded vs. error if required?
        status = Status.ok

      logger.debug('runplaybook status %s results %s', status, results.results[0]._check_key('result'))

      applied = len(results.resultsByStatus.ok) + len(results.resultsByStatus.failed)
      # XXX if more then one task??
      result = results.results and ansibleResults(results.results[0], self.getResultKeys(task, results.results[0])) or None
      if result and status == Status.ok or status == Status.degraded:
        # this can update resources so don't do it on error
        self._processResult(task, result)
      yield task.createResult(applied > 0, results.changed > 0, status, result=result)
    finally:
      self._cleanup()

# see https://github.com/ansible/ansible/blob/d72587084b4c43746cdb13abb262acf920079865/examples/scripts/uptime.py
# and https://github.com/ansible/ansible/blob/devel/lib/ansible/plugins/callback/log_plays.py
_ResultsByStatus = collections.namedtuple('_ResultsByStatus', "ok failed skipped unreachable")
class ResultCallback(CallbackModule):
  # NOTE: callbacks will run in seperate process
  #see ansible.executor.task_result.TaskResult and ansible.playbook.task.Task

  def __init__(self):
    super(ResultCallback, self).__init__()
    self.results = []
    # named tuple of OrderedDict<task_name:list<result>>
    self.resultsByStatus = _ResultsByStatus(*[collections.OrderedDict() for x in range(4)])
    self._load_name = 'result'
    self.changed = 0

  def getInfo(self, result):
    host = result._host
    taskname = result.task_name
    fields = result._task_fields.keys()
    keys = result._result.keys()
    return "%s: %s(%s) => %s" % (host, taskname, fields, keys)

  def _addResult(self, status, result):
    self.results.append(result)
    if result._result.get('changed', False):
      self.changed += 1
    # XXX should save by host too
    getattr(self.resultsByStatus, status).setdefault(result.task_name, []).append(result)

  def v2_runner_on_ok(self, result):
    self._addResult('ok', result)
    #print("ok", self.getInfo(result))
    super(ResultCallback, self).v2_runner_on_ok(result)

  def v2_runner_on_skipped(self, result):
    self._addResult('skipped', result)
    #print("skipped", self.getInfo(result))
    super(ResultCallback, self).v2_runner_on_skipped(result)

  def v2_runner_on_failed(self, result, **kwargs):
    self._addResult('failed', result)
    #print("failed", self.getInfo(result))
    super(ResultCallback, self).v2_runner_on_failed(result, **kwargs)

  def v2_runner_on_unreachable(self, result):
    self._addResult('unreachable', result)
    #print("unreachable", self.getInfo(result))
    super(ResultCallback, self).v2_runner_on_unreachable(result)

def runPlaybooks(playbooks, _inventory, params=None, args=None):
  # unfurl.util should have initialized ansibleDummyCli and ansibleDisplay already
  inventoryArgs =  ['-i', _inventory] if _inventory else []
  args=['ansible-playbook'] + inventoryArgs + (args or []) + playbooks
  logger.info('running ' + ' '.join(args))
  cli = PlaybookCLI(args)
  cli.parse()

  # CallbackBase imports __main__.cli (which is set to ansibleDummyCli)
  # as assigns its options to self._options
  ansibleDummyCli.options.__dict__.update(cli.options.__dict__)

  # replace C.DEFAULT_STDOUT_CALLBACK with our own so we have control over logging
  # config/base.yml sets C.DEFAULT_STDOUT_CALLBACK == 'default' (ansible/plugins/callback/default.py)
  # (cli/console.py and cli/adhoc.py sets it to 'minimal' but PlaybookCLI.run() in cli/playbook.py uses the default)
  #see also https://github.com/projectatomic/atomic-host-tests/blob/master/callback_plugins/default.py
  resultsCB = ResultCallback()
  resultsCB.set_options()
  C.DEFAULT_STDOUT_CALLBACK = resultsCB

  _play_prereqs = cli._play_prereqs
  def hook_play_prereqs(options):
    loader, inventory, variable_manager = _play_prereqs(options)
    if params:
      variable_manager._extra_vars.update(params)
    resultsCB.inventoryManager = inventory
    resultsCB.variableManager = variable_manager
    # XXX inventory.localhost is None right now
    # inventory.localhost.set_variable("ansible_python_interpreter", sys.executable)
    return loader, inventory, variable_manager
  cli._play_prereqs = hook_play_prereqs

  oldVerbosity = ansibleDisplay.verbosity
  if logging.getLogger('ansible').getEffectiveLevel() <= 10: # debug
    ansibleDisplay.verbosity = 2
  try:
    if cli.options.verbosity > ansibleDisplay.verbosity:
      ansibleDisplay.verbosity = cli.options.verbosity
    resultsCB.exit_code = cli.run()
  finally:
    ansibleDisplay.verbosity = oldVerbosity
  return resultsCB



from collections import Mapping, MutableSequence, MutableMapping
from .util import diffDicts

def serializeValue(value, **kw):
  getter = getattr(value, 'asRef', None)
  if getter:
    return getter(kw)
  if isinstance(value, Mapping):
    return dict((key, serializeValue(v, **kw)) for key, v in value.items())
  elif isinstance(value, (MutableSequence, tuple)):
    return [serializeValue(item, **kw) for item in value]
  else:
    return value

class ResourceRef(object):
  # ABC requires 'parent', and '_resolve'

  def _getProp(self, name):
    if name == '.':
      return self
    elif name == '..':
      return self.parent
    name = name[1:]
    # XXX3 use propmap
    return getattr(self, name)

  def __reflookup__(self, key):
    if not key:
      raise KeyError(key)
    if key[0] == '.':
      return self._getProp(key)

    return self._resolve(key)

  def yieldParents(self):
    "yield self and ancestors starting from self"
    resource = self
    while resource:
      yield resource
      resource = resource.parent

  @property
  def ancestors(self):
    return list(self.yieldParents())

  @property
  def parents(self):
    """list of parents starting from root"""
    return list(reversed(self.ancestors))[:-1]

  @property
  def root(self):
    return self.ancestors[-1]

  @property
  def all(self):
    return self.root._all

  @property
  def templar(self):
    return self.root._templar

class ChangeRecord(object):
  def __init__(self, changeId=0, parentId=None, commitId='', startTime=''):
    self.changeId = changeId
    self.parentId = parentId
    self.commitId = commitId
    self.startTime = startTime

class ChangeAware(object):
  def hasChanged(self, changeRecord):
    return False

class ExternalValue(ChangeAware):
  __slots__ = ('type', 'key')

  def __init__(self, type, key):
    self.type = type
    self.key = key

  def get(self):
    return self.key

# XXX def __setstate__

  def __eq__(self, other):
    if isinstance(other, ExternalValue):
      return self.get() == other.get()
    return self.get() == other

  def resolveKey(self, key=None, currentResource=None):
    if key:
      value = self.get()
      getter = getattr(value, '__reflookup__', None)
      if getter:
        return getter(key)
      else:
        return value[key]
    else:
      return self.get()

  def asRef(self, options=None):
    if options and options.get('resolveExternal'):
      return serializeValue(self.get(), **options)
    serialized = {self.type: self.key}
    return {'eval': serialized}

_Deleted = object()
class Result(ChangeAware):
  __slots__ = ('original', 'resolved', 'external', 'select')

  def __init__(self, resolved, original = _Deleted):
    self.select = ()
    self.original = original
    if isinstance(resolved, ExternalValue):
      self.resolved = resolved.get()
      assert not isinstance(self.resolved, Result), self.resolved
      self.external = resolved
    else:
      assert not isinstance(resolved, Result), resolved
      self.resolved = resolved
      self.external = None

  def asRef(self, options=None):
    if self.external:
      ref = self.external.asRef()
      if self.select:
        ref['foreach'] = "." + "::".join(self.select)
      return ref
    else:
      val = serializeValue(self.resolved, **(options or {}))
      return val

  def hasDiff(self):
    if self.original is _Deleted: # this is a new item
      return True
    else:
      if isinstance(self.resolved, Results):
        return self.resolved.hasDiff()
      else:
        newval = self.asRef()
        if self.original != newval:
          return True
    return False

  def getDiff(self):
    if isinstance(self.resolved, Results):
      return self.resolved.getDiff()
    else:
      val = self.asRef()
      if not self.external and isinstance(val, Mapping):
        old = serializeValue(self.original)
        if isinstance(old, Mapping):
          return diffDicts(old, val)
      return val

  def _values(self):
    resolved = self.resolved
    if isinstance(resolved, ResultsList):
      # iterate on list to make sure __getitem__ was called
      return (resolved._attributes[i] for (i,v) in enumerate(resolved))
    elif isinstance(resolved, ResultsMap):
      # use items() to make sure __getitem__ was called
      return (resolved._attributes[k] for (k, v) in resolved.items())
    elif isinstance(resolved, Mapping):
      return resolved.values()
    elif isinstance(resolved, MutableSequence):
      return (Result(i) for i in resolved)
    else:
      return resolved

  def _resolveKey(self, key, currentResource):
    # might return a Result
    if self.external:
      value = self.external.resolveKey(key, currentResource)
    else:
      getter = getattr(self.resolved, '__reflookup__', None)
      if getter:
        value = getter(key)
      else:
        value = self.resolved[key]
    return value

  def project(self, key, ctx):
    # returns a Result
    from .eval import Ref
    value = self._resolveKey(key, ctx._lastResource)
    if isinstance(value, Result):
      result = value
    elif Ref.isRef(value):
      result = Ref(value).resolve(ctx, wantList = 'result')
      if not result:
        raise KeyError(key)
    else:
      result = Result(value)
    if self.external:
      # if value is an ExternalValue this will overwrite it
      result.external = self.external
      result.select = self.select + (key,)
    return result

  def hasChanged(self, changeset):
    if self.external:
      return self.external.hasChanged(changeset)
    elif isinstance(self.resolved, ChangeAware):
      return self.resolved.hasChanged(changeset)
    else:
      return False

  def __eq__(self, other):
    if isinstance(other, Result):
      return self.resolved == other.resolved
    else:
      return self.resolved == other

  def __repr__(self):
    return "Result(%r, %r, %r)" % (self.resolved, self.external, self.select)

class Results(object):
  """
  Evaluating expressions are not guaranteed to be idempotent (consider quoting)
  and resolving the whole tree up front can lead to evaluations of circular references unless the
  order is carefully chosen. So evaluate lazily and memoize the results.
  This also allows us to track changes to the returned structure.
  """

  __slots__ = ('_attributes', 'context', '_deleted')

  doFullResolve=False
  def __init__(self, serializedOriginal, resourceOrCxt):
      from .eval import RefContext
      assert not isinstance(serializedOriginal, Results), serializedOriginal
      self._attributes = serializedOriginal
      self._deleted = {}
      if not isinstance(resourceOrCxt, RefContext):
        ctx = RefContext(resourceOrCxt)
      else:
        ctx = resourceOrCxt

      oldBaseDir = ctx.baseDir
      newBaseDir = getattr(serializedOriginal, 'baseDir', oldBaseDir)
      if newBaseDir and newBaseDir != oldBaseDir:
        ctx = ctx.copy()
        ctx.baseDir = newBaseDir
        ctx.trace("found baseDir", newBaseDir, 'old', oldBaseDir)
      self.context = ctx

  def hasDiff(self):
    return any(isinstance(x, Result) and x.hasDiff() for x in self._attributes)

  def getCopy(self, key, default=None):
    from .eval import mapValue
    try:
      val = self._attributes[key]
    except:
      val = default
    else:
      if isinstance(val, Result):
        assert not isinstance(val.resolved, Result), val
        if val.original is _Deleted:
          val = val.asRef()
        else:
          val = val.original
    return mapValue(val, self.context)

  @staticmethod
  def _mapValue(val, context):
    "Recursively and lazily resolves any references in a value"
    from .eval import mapValue, Ref
    if isinstance(val, Results):
      return val
    elif Ref.isRef(val):
      return Ref(val).resolve(context, wantList = 'result')
    elif isinstance(val, Mapping):
      return ResultsMap(val, context)
    elif isinstance(val, (list,tuple)):
      return ResultsList(val, context)
    else:
      # at this point, just evaluates templates in strings or returns val
      return mapValue(val, context)

  def __getitem__(self, key):
    from .eval import mapValue
    val = self._attributes[key]
    if isinstance(val, Result):
      assert not isinstance(val.resolved, Result), val
      return val.resolved
    else:
      if self.doFullResolve:
        resolved = mapValue(val, self.context)
      else:
        self.context.trace('Results._mapValue', val)
        resolved = self._mapValue(val, self.context)
      # will return a Result if val was an expression that was evaluated
      if isinstance(resolved, Result):
        result = resolved
        result.original = val
        resolved = result.resolved
      else:
        result = Result(resolved, val)
      self._attributes[key] = result
      assert not isinstance(resolved, Result), val
      return resolved

  def __setitem__(self, key, value):
    assert not isinstance(value, Result), (key, value)
    self._attributes[key] = Result(value)
    self._deleted.pop(key, None)

  def __delitem__(self, index):
    val = self._attributes[index]
    self._deleted[index] = val
    del self._attributes[index]

  def __len__(self):
    return len(self._attributes)

  def __eq__(self, other):
    if isinstance(other, Results):
      return self._attributes == other._attributes
    else:
      self.resolveAll()
      return self._attributes == other

  def __repr__(self):
    return "Results(%r)" % self._attributes


class ResultsMap(Results, MutableMapping):

  def __iter__(self):
      return iter(self._attributes)

  def resolveAll(self):
    list(self.values())

  def getDiff(self, cls=dict):
    # returns a dict with the same semantics as diffDicts
    diffDict = cls()
    for key, val in self._attributes.items():
      if isinstance(val, Result) and val.hasDiff():
        diffDict[key] = val.getDiff()

    for key in self._deleted:
      diffDict[key]= {'+%': 'delete'}

    return diffDict


class ResultsList(Results, MutableSequence):

  def insert(self, index, value):
    assert not isinstance(value, Result), value
    self._attributes.insert(index, Result(value))

  def getDiff(self, cls=list):
    # we don't have patchList yet so just returns the whole list
    return cls(val.getDiff() if isinstance(val, Result) else val for val in self._attributes)

  def resolveAll(self):
    list(self)



"""
Project folder structure:

project/spec/.git # has service-template.yaml and manifest-template.yaml
        instances/current/.git # has manifest.yaml
        # subprojects are created by repository declarations in spec or instance
        subproject/spec/
                  instances/current
        # unfurl.yaml merges with ~/.unfurl_home/unfurl.yaml
        unfurl.yaml # might create 'secret' or 'local' subprojects
        revisions/...
"""
import uuid
import os
import os.path
from git import Repo
from . import __version__
from .tosca import TOSCA_VERSION

def writeLocalConfig(projectdir):
  os.makedirs(projectdir)
  filepath = os.path.join(projectdir, 'unfurl.yaml')
  with open(filepath, 'w') as f:
    f.write("""\
unfurl:
  version: %s
instances:
  - file: instances/current/manifest.yaml
    default: true

# this is the default behavior, so not needed:
# defaults: # used if the instance isn't defined above
#   local:
#     # local and secret can have "attributes" instead of declaring an import
#     attributes:
#       inheritFrom: home
#   secret:
#     attributes:
#       inheritFrom: home
""" % __version__)
  return filepath

def createHome(path=None):
  """
  Write ~/.unfurl_home/unfurl.yaml if missing
  """
  homedir = path or os.path.expanduser(os.path.join('~', '.unfurl_home'))
  if not os.path.exists(homedir):
    return writeLocalConfig(homedir)
  # XXX when is unfurl.yaml instances deployed?

def createRepo(repotype, gitDir, gitUri=None):
  os.makedirs(gitDir)
  repo = Repo.init( gitDir )
  filename = '.unfurl'
  filepath = os.path.join(gitDir, filename)
  with open(filepath, 'w') as f:
    f.write("""\
  unfurl:
    version: %s
  repo:
    type: %s
    uuid: %s
  """ % (__version__, repotype, uuid.uuid1()))

  repo.index.add([filename])
  repo.index.commit("Initial Commit")
  return repo

def createSpecRepo(gitDir):
  repo = createRepo('spec', gitDir)
  serviceTemplatePath = os.path.join(gitDir, 'service-template.yaml')
  with open(serviceTemplatePath, 'w') as f:
    f.write("""\
tosca_definitions_version: %s
repositories:
  spec:
    url: file:.
    metadata:
      initial-commit: %s
topology_template:
  node_templates:
""" % (TOSCA_VERSION, repo.head.commit.hexsha))
  manifestTemplatePath = os.path.join(gitDir, 'manifest-template.yaml')
  with open(manifestTemplatePath, 'w') as f:
    f.write("""\
  apiVersion: unfurl/v1alpha1
  kind: Manifest
  spec:
    tosca:
      +%include: service-template.yaml
  status: {}
      """)
  repo.index.add(['service-template.yaml', 'manifest-template.yaml'])
  repo.index.commit("Default specification repository boilerplate")
  return repo

def createInstanceRepo(gitDir, specRepo):
  repo = createRepo('instance', gitDir)
  filepath = os.path.join(gitDir, 'manifest.yaml')
  relPathToSpecRepo = os.path.relpath(specRepo.working_tree_dir, os.path.abspath(gitDir))
  specInitialCommit = list(specRepo.iter_commits('HEAD', max_parents=0))[0].hexsha
  with open(filepath, 'w') as f:
    f.write("""\
apiVersion: unfurl/v1alpha1
kind: Manifest
# merge in manifest-template.yaml from spec repo
+%%include:
  file: manifest-template.yaml
  repository: spec
spec:
  tosca:
    repositories:
      spec:
        url: file:%s
        metadata:
          initial-commit: %s
      instance:
        url: file:.
        metadata:
          initial-commit: %s
""" % (relPathToSpecRepo, specInitialCommit, repo.head.commit.hexsha))
  repo.index.add(['manifest.yaml'])
  repo.index.commit("Default instance repository boilerplate")
  return repo

def createProject(projectdir, home=None):
  """
  # creates .unfurl, project/specs/manifest-template.yaml, instances/current/manifest.yaml
  # init git repos with initial commits
  # adds ~/.unfurl_home/unfurl.yaml if missing
  # add project to ~/.unfurl_home/unfurl.yaml
  """
  newHome = createHome(home)
  projectConfigPath = writeLocalConfig(projectdir)
  specRepo = createSpecRepo(os.path.join(projectdir, 'spec'))
  createInstanceRepo(os.path.join(projectdir, 'instances', 'current'), specRepo)
  return newHome, projectConfigPath

def createNewInstance(specRepoDir, targetPath):
  from .repo import Repo
  sourceRepo = Repo.createGitRepoIfExists(specRepoDir)
  if not sourceRepo:
    return None, "No repository exists at " + os.path.abspath(specRepoDir)
  if not sourceRepo.isValidSpecRepo():
    return None, "The respository at '%s' is not valid" % os.path.abspath(specRepoDir)

  # XXX
  #if localEnv.findPathInRepos(targetPath):
  #  return None # "can't create repo in another repo"
  instanceRepo = createInstanceRepo(targetPath, sourceRepo.repo)
  # XXX
  #project = localEnv.findProject(targetPath)
  #if project: # add to project
  return instanceRepo, "created new instance repository at %s" % os.path.abspath(targetPath)

def cloneSpecToNewProject(sourceDir, projectDir):
  from .repo import Repo
  sourceRepo = Repo.createGitRepoIfExists(sourceDir)
  if not sourceRepo:
    return None, "No repository exists at " + os.path.abspath(sourceDir)
  if not sourceRepo.isValidSpecRepo():
    return None, "The respository at '%s' is not valid" % os.path.abspath(sourceDir)

  if os.path.exists(projectDir):
    return None, os.path.abspath(projectDir) + ": file already exists"

  # XXX make sure projectdir is usable
  projectConfigPath = writeLocalConfig(projectDir)
  fullProjectDir = os.path.abspath(projectDir)
  specRepo = sourceRepo.clone(os.path.join(fullProjectDir, 'spec'))
  createInstanceRepo(os.path.join(projectDir, 'instances', 'current'), specRepo.repo)
  return projectConfigPath, "New project created at %s" % fullProjectDir



import unittest
from unfurl.yamlmanifest import YamlManifest
from unfurl.job import Runner, JobOptions
from unfurl.support import Status
from unfurl.configurator import Configurator
# from unfurl.util import UnfurlError, UnfurlValidationError

class SetAttributeConfigurator(Configurator):
  def run(self, task):
    task.target.attributes['private_address'] = '10.0.0.1'
    yield task.createResult(True, True, Status.ok)

manifestDoc = '''
apiVersion: unfurl/v1alpha1
kind: Manifest
spec:
  inputs:
    cpus: 2
  tosca:
    node_types:
      testy.nodes.aNodeType:
        derived_from: tosca.nodes.Root
        properties:
          private_address:
            type: string
            metadata:
              sensitive: true
    topology_template:
      inputs:
        cpus:
          type: integer
          description: Number of CPUs for the server.
          constraints:
            - valid_values: [ 1, 2, 4, 8 ]
          metadata:
            sensitive: true
      outputs:
        server_ip:
          description: The private IP address of the provisioned server.
          # equivalent to { get_attribute: [ my_server, private_address ] }
          value: {eval: "::my_server::private_address"}
      node_templates:
        testSensitive:
          type: testy.nodes.aNodeType
          properties:
            private_address: foo
          interfaces:
           Standard:
            create:
              implementation:
                primary: SetAttributeConfigurator
        my_server:
          type: tosca.nodes.Compute
          properties:
            test:  { concat: ['cpus: ', {get_input: cpus }] }
          capabilities:
            # Host container properties
            host:
             properties:
               num_cpus: { eval: ::inputs::cpus }
               disk_size: 10 GB
               mem_size: 512 MB
            # Guest Operating System properties
            os:
              properties:
                # host Operating System image properties
                architecture: x86_64
                type: Linux
                distribution: RHEL
                version: 6.5
          interfaces:
           Standard:
            create:
              inputs:
                inputs:
                priority:
                inputSchema:
              implementation:
                primary: SetAttributeConfigurator
                timeout: 120
'''

class ToscaSyntaxTest(unittest.TestCase):
  def test_inputAndOutputs(self):
    manifest = YamlManifest(manifestDoc)
    job = Runner(manifest).run(JobOptions(add=True, startTime="time-to-test"))
    assert not job.unexpectedAbort, job.unexpectedAbort.getStackTrace()
    assert manifest.getRootResource().findResource('my_server').attributes['test'], 'cpus: 2'
    assert job.getOutputs()['server_ip'], '10.0.0.1'
    assert job.status == Status.ok, job.summary()
    # XXX verify redacted output
    # print(job.out.getvalue())



import six
from .runtime import Resource
from .util import UnfurlError, lookupClass
from .support import Status
from .configurator import ConfigurationSpec

import logging
logger = logging.getLogger('unfurl')

class Plan(object):
  """
  create:  template or unapplied resource
  upgrade: resource
  delete:  resource
  check:   resource

  options:
  --append with create to avoid error if exists
  in the future, should run with previous command on this resource or template
  use:configurator use that configurator
  """
  rootConfigurator = None # XXX3

  def __init__(self, root, toscaSpec, jobOptions):
    self.jobOptions = jobOptions
    self.root = root
    self.tosca = toscaSpec
    assert self.tosca

  def findResourcesFromTemplate(self, nodeTemplate):
    for resource in self.root.getSelfAndDescendents():
      if resource.template.name == nodeTemplate.name:
        yield resource

  def findParentResource(self, source):
    parentTemplate = findParentTemplate(source.toscaEntityTemplate)
    if not parentTemplate:
      return self.root
    for parent in self.findResourcesFromTemplate(parentTemplate):
      # XXX need to evaluate matches
      return parent

  def createResource(self, template):
    # XXX create capabilities and requirements too?
    # XXX if requirement with HostedOn relationshio, target is the parent not root
    parent = self.findParentResource(template)
    assert parent, "parent should have already been created"
    return Resource(template.name, template=template, parent=parent)

  def createShellConfigurator(self, cmdLine, action, inputs=None, timeout=None):
      params = dict(command=cmdLine, timeout=timeout)
      if inputs:
        params.update(inputs)
      return ConfigurationSpec('cmdline', action, className='unfurl.shellconfigurator.ShellConfigurator', inputs = params)

  def getConfigurationSpecFromInterface(self, iDef):
    '''implementation can either be a named artifact (including a python configurator class),
      configurator node template, or a file path'''
    implementation = iDef.implementation
    timeout = None
    if isinstance(implementation, dict):
      timeout = implementation.get('timeout')
      implementation = implementation.get('primary')
      if isinstance(implementation, dict):
        # it's an artifact definition
        # XXX retrieve from repository if defined
        implementation = implementation.get('file')
    if not implementation or not isinstance(implementation, six.string_types):
      raise UnfurlError('invalid implementation spec %s' % implementation)
    configuratorTemplate = self.tosca.configurators.get(implementation)
    if configuratorTemplate:
      attributes = configuratorTemplate.properties
      kw = { k : attributes[k] for k in set(attributes) & set(ConfigurationSpec.getDefaults()) }
      if 'inputs' not in kw:
        kw['inputs'] = iDef.inputs
      if 'className' not in kw:
        kw['className'] = configuratorTemplate.getInterfaces()[0].implementation
      return ConfigurationSpec(configuratorTemplate.name, iDef.name, **kw)
    else:
      try:
        lookupClass(implementation)
        return ConfigurationSpec(implementation, iDef.name, className=implementation, inputs = iDef.inputs)
      except UnfurlError:
        # assume its executable file, create a ShellConfigurator
        return self.createShellConfigurator([implementation], iDef.name, iDef.inputs, timeout=timeout)

  def findImplementation(self, interfaceType, operation, template):
    for iDef in template.getInterfaces():
      if iDef.type == interfaceType and iDef.name == operation:
        return self.getConfigurationSpecFromInterface(iDef)
    return None

  def includeTask(self, template, resource, oldTemplate):
    """ Returns whether or not the config should be included in the current job.

Reasons include: "all", "add", "upgrade", "update", "re-add", 'revert obsolete',
'never applied', "config changed", "failed to apply", "degraded", "error".

Args:
    config (ConfigurationSpec): The :class:`ConfigurationSpec` candidate
    lastChange (Configuration): The :class:`Configuration` representing the that last time
      the given :class:`ConfigurationSpec` was applied or `None`

Returns:
    (str, ConfigurationSpec): Returns a pair with reason why the task was included
      and the :class:`ConfigurationSpec` to run or `None` if it shound't be included.
    """
    jobOptions = self.jobOptions
    assert resource
    if jobOptions.all and template:
      return 'all', template
    if template and resource.status == Status.notapplied and not resource.lastConfigChange and jobOptions.add:
        return 'add', template

    if not template:
      if jobOptions.revertObsolete:
        return 'revert obsolete', oldTemplate
      if jobOptions.all:
        return 'all', oldTemplate
      if resource.status == Status.notapplied and jobOptions.add:
        return 'never applied', oldTemplate
    elif template != oldTemplate:
      # the user changed the configuration:
      if jobOptions.upgrade:
        return 'upgrade', template
      if resource.status == Status.notpresent and jobOptions.add:
        # this case is essentially a re-added config, so re-run it
        return 're-add', template
      if jobOptions.update:
        # apply the new configuration unless it will trigger a major version change
        if False:  # XXX if isMinorDifference(template, oldTemplate)
          return 'update', template

    # there isn't a new config to run, see if the last applied config needs to be re-run
    # XXX: if (jobOptions.upgrade or jobOptions.update or jobOptions.all):
    #  if (configTask.hasInputsChanged() or configTask.hasDependenciesChanged()) and
    #    return 'config changed', configTask.configSpec
    return self.checkForRepair(resource, oldTemplate)

  def checkForRepair(self, instance, lastTemplate):
    jobOptions = self.jobOptions

    assert instance
    # spec = lastChange.configurationSpec

    if jobOptions.repair=="none":
      return None
    status = instance.status
    if status == Status.notapplied and instance.required:
      status = Status.error # treat as error

    # repair should only apply to configurations that are active and in need of repair
    # XXX2 what about pending??
    if status not in [Status.degraded, Status.error, Status.notapplied]:
      return None

    if status == Status.notapplied:
      if jobOptions.repair == 'notapplied':
        return 'failed to apply', lastTemplate
      else:
        return None
    if jobOptions.repair == "degraded":
      assert status > Status.ok, status
      return 'degraded', lastTemplate # repair this
    elif status == Status.degraded:
      assert jobOptions.repair == 'error', jobOptions.repair
      return None # skip repairing this
    else:
      assert jobOptions.repair == 'error', "repair: %s status: %s" % (jobOptions.repair, instance.status)
      return 'error', lastTemplate # repair this

  def executePlan(self):
    """
    Find candidate tasks

    Given declared spec, current status, and job options, generate selector

    does the config apply to the action?
    is it out of date?
    is it in a ok state?
    has its configuration changed?
    has its dependencies changed?
    are the resources it modifies in need of repair?
    manual override (include / skip)

    yields configSpec, target, reason
    """
    opts = self.jobOptions
    templates = [] if not self.tosca.nodeTemplates else [
      t for t in self.tosca.nodeTemplates.values()
        if not t.isCompatibleType(self.tosca.ConfiguratorType)]

    if opts.template:
      filterTemplate = self.tosca.getTemplate(opts.template)
      if not filterTemplate:
        raise UnfurlError('specified template not found %s' % filterTemplate)
    else:
      filterTemplate = None

    # order by ancestors
    templates = list(orderTemplates(
        self.tosca.template.topology_template.graph,
        {t.name : t for t in templates}, filterTemplate and filterTemplate.name))

    logger.debug('checking for tasks for templates %s', [t.name for t in templates])
    visited = set()
    for template in templates:
      found = False
      for resource in self.findResourcesFromTemplate(template):
        found = True
        visited.add(id(resource))
        include = self.includeTask(template, resource, resource.template)
        if include:
          reason, template = include
          if resource.status.notapplied or resource.status.notpresent:
            operation = 'create'
          else:
            operation = 'configure'
          yield self.generateConfiguration(operation, resource, reason, opts.useConfigurator)
        else:
          logger.debug('skipping task for %s:%s', resource.name, template.name),

      if not found and (opts.add or opts.all):
        reason = 'add'
        operation = 'create'
        # XXX create NodeInstance instead to include relationships
        resource = self.createResource(template)
        visited.add(id(resource))
        yield self.generateConfiguration(operation, resource, reason, opts.useConfigurator)

    if opts.revertObsolete: #XXX expose option in cli (as --prune ?)
      for instance in self.root.getOperationalDependencies():
        for resource in instance.getSelfAndDescendents():
          if id(resource) not in visited:
            logger.debug('checking for tasks for orphaned resource %s', resource.name)
            # it's an orphaned config
            include = self.includeTask(None, resource, resource.template)
            if not include:
              continue
            reason, config = include
            yield self.generateConfiguration('delete', resource, reason)
    # #XXX opts.create, opts.append, opts.cmdline, opts.useConfigurator

  def generateConfiguration(self, action, resource, reason=None, cmdLine=None, useConfigurator=None):
    # XXX update joboptions, useConfigurator
    if cmdLine:
      configSpec = self.createShellConfigurator(cmdLine, action)
    else:
      configSpec = self.findImplementation('Standard', action, resource.template)
    if not configSpec:
      raise UnfurlError('unable to find an implementation to "%s" "%s" on ""%s"' % (action, resource.template.name, resource.template.name) )
    logger.debug('creating configuration %s with %s to run for %s: %s', configSpec.name, configSpec.inputs, resource.name, reason or action)
    return (configSpec, resource, reason or action)

def orderTemplates(graph, templates, filter=None):
  seen = set()
  for source in graph:
    if filter and source.name != filter:
      continue
    if source in seen:
      continue
    for ancestor in getAncestorTemplates(source):
      if ancestor in seen:
        continue
      seen.add(ancestor)
      template = templates.get(ancestor.name)
      if template:
        yield template

def getAncestorTemplates(source):
  for target, relation in source.related.items():
    if relation.type == 'tosca.relationships.HostedOn':
      for ancestor in getAncestorTemplates(target):
        yield ancestor
      break
  yield source

def findParentTemplate(source):
  for target, relation in source.related.items():
    if relation.type == 'tosca.relationships.HostedOn':
      return target
    return None


# XXX!:
def buildDependencyGraph():
  """
  We need to find each executed configuration that is affected by a configuration change
  and re-execute them

  dependencies map to inbound edges
  lastConfigChange filters ConfigTasks
  ConfigTasks.configurationResource dependencies are inbound
  keys in ConfigTasks.changes map to a outbound edges to resources

  Need to distinguish between runtime dependencies and configuration dependencies?
  """

def buildConfigChangedExecutionPlan():
  """
  graph = buildDependencyGraph()

  We follow the edges if a resource's attributes or dependencies have changed
  First we validate that we can execute the plan by validating configurationResource in the graph
  If it validates, add to plan

  After executing a task, only follow the dependent outputs that changed no need to follow those dependencies
  """

def validateNode(resource, expectedTemplateOrType):
  """
  First make sure the resource conforms to the expected template (type, properties, attributes, capabilities, requirements, etc.)
  Then for each dependency that corresponds to a required relationship, call validateNode() on those resources with the declared type or template for the relationship
  """

def buildCreateExecutionPlan(self):
  """
  move Configurator nodes to separate list
  start with root node template,
  configuratorTemplate = find Configurator node for creation,
  self.add(self.buildCreateExecutionPlan(configuratorTemplate))
  return # that's all we need, the rest is dynamic, for each new resource, find Configurator for every missing capability and requirement
  """

def buildCreateStaticExecutionPlan(self):
  """
  Build a static plan by using the 'provides' property to estimate what the
  configuration will create and then recursively find missing capabilities and requirements and then the configurators to run
  """
  start = self.buildCreateExecutionPlan(self.rootNodeTemplate)

def buildUpgradeExecutionPlan():
  """
  Same as buildCreateExecutionPlan except look for existing resources
  if it exists, see if it needs upgrading, if it doesn't then find create configurator
  Upgrading:
    compare resource with spec,
    if different add configurators to replace exisiting configuration
    for each requirement not in dependencies, add configurator
    for each dependency on requirement, see if corresponding requirement in spec
      if not, remove dependency
      if it is, call buildUpgradeExecutionPlan on the target resource with the target node type or template
  """

def buildUpdateExecutionPlan():
  """
  Only apply updates that don't change the currently applied spec,
  Starting with the start resource compare deployed artifacts and software nodes associate with it with current template
  and if diffence is no more than a minor version bump,
  retreive the old version of the topology that is associated with the appropriate configuredBy
  and with it try to find and queue a configurator that can apply those changes.

  For each resource dependency, call buildUpdateExecutionPlan().
  """



"""Loads and saves a Unfurl manifest with the following format:

.. code-block:: YAML

  apiVersion: VERSION
  kind: Manifest
  imports:
    name: # manifest to represent as a resource
      file: # if is missing, manifest must declared in local config
      repository:
      resource: name # default is root
      attributes: # queries into manifest
      properties: # expected JSON schema for attributes
    # locals and secrets are special cases that must be defined in local config
    locals:
     properties:
       name1:
         type: string
         default
     required:
    secrets:
     properties:
       name1:
         type: string
         default
       # save-digest: true
       # prompt: true
       # readonly: true
     required:

  spec:
    tosca:
      # <tosca service template>
    inputs:
    instances:
      name:
        template:
        implementations:
          action: implementationName
    implementations:
      name1:
        className
        version
        intent
        inputs:
  status:
    topology: repo:topology_template:revision
    inputs:
    outputs:
    readyState: #, etc.
    instances:
      name: # root resource is always named "root"
        template: repo:templateName:revision
        attributes:
          .interfaces:
            interfaceName: foo.bar.ClassName
        capabilities:
          - name:
            attributes:
            operatingStatus
            lastChange: changeId
        requirements:
         - name:
           operatingStatus
           lastChange
        readyState:
          effective:
          local:
          lastConfigChange: changeId
          lastStateChange: changeId
        priority
      resources:
        child1:
          # ...
  repositories
  changeLog: changes.yaml
  latestChanges:
    jobId: 1
    startCommit: ''
    startTime:
    specDigest:
    lastChangeId:
    readyState:
      effective: error
      local: ok

# in changes.yaml:
  - jobId: 1
    startCommit:
    endCommit: ''
    startTime:
    specDigest:
    tasksRun:
    readyState:
      effective: error
      local: ok
  - changeId: 2
    startTime
    parentId: 1 # allows execution plan order to be reconstructed
    previousId # XXX last time this configuration ran
    target
    readyState
    priority
    resource
    config
    action
    implementation:
      type: resource | artifact | class
      key: repo:key#commitid | className:version
    inputs
    dependencies:
      - ref: ::resource1::key[~$val]
        expected: "value"
      - name: named1
        ref: .configurations::foo[.operational]
        required: true
        schema:
          type: array
    changes:
      resource1:
        .added: # set if added resource
        .status: # set when adding or removing
        foo: bar
      resource2:
        .spec:
        .status: notpresent
      resource3/child1: +%delete
    messages: []
"""

import six
import sys
import collections
import numbers
import os.path
import itertools

from .util import (UnfurlError, VERSION, toYamlText, restoreIncludes, patchDict)
from .yamlloader import YamlConfig, loadFromRepo, load_yaml, yaml
from .result import serializeValue
from .support import ResourceChanges, Status, Priority, Action
from .localenv import LocalEnv
from .job import JobOptions, Runner
from .manifest import Manifest, ChangeRecordAttributes
from .runtime import TopologyResource, Resource

from ruamel.yaml.comments import CommentedMap
from codecs import open

import logging
logger = logging.getLogger('unfurl')

# XXX3 add as file to package data
#schema=open(os.path.join(os.path.dirname(__file__), 'manifest-v1alpha1.json')).read()
schema = {
  "$schema": "http://json-schema.org/draft-04/schema#",
  "$id": "https://www.onecommons.org/schemas/unfurl/v1alpha1.json",
  "definitions": {
    "atomic": {
      "type": "object",
      "properties": {
        # XXX
        # "+%": {
        #   "default": "replaceProps"
        # }
      },
    },
    "namedObjects": {
      "type": "object",
      "propertyNames": {
          "pattern": r"^[A-Za-z_][A-Za-z0-9_\-]*$"
        },
    },
    "secret": {
      "type": "object",
      "properties": {
        "ref": {
          "type": "object",
          "properties": {
            "secret": {
              "type": "string",
              "pattern": r"^[A-Za-z_][A-Za-z0-9_\-]*$"
            },
          },
          "required": ["secret"]
        },
      },
      "required": ["ref"]
    },
    "attributes": {
      "allOf": [
        { "$ref": "#/definitions/namedObjects" },
        { "$ref": "#/definitions/atomic" }
      ],
      'default': {}
    },
    "resource": {
      "type": "object",
      "allOf": [
        { "$ref": "#/definitions/status" },
        {"properties": {
          "template": {"type": "string"},
          "attributes": {
            "$ref": "#/definitions/attributes",
            "default": {}
           },
          "resources": {
            "allOf": [
              { "$ref": "#/definitions/namedObjects" },
              {"additionalProperties": { "$ref": "#/definitions/resource" }}
            ],
            'default': {}
          },
        }}
      ]
     },
    "configurationSpec": {
      "type": "object",
      "properties": {
        "className": {"type":"string"},
        "majorVersion": {"anyOf": [{"type":"string"}, {"type":"number"}]},
        "minorVersion": {"type":"string"},
        "intent": { "enum": list(Action.__members__) },
        "inputs": {
          "$ref": "#/definitions/attributes",
          "default": {}
         },
        "preconditions": {
          "$ref": "#/definitions/schema",
          "default": {}
         },
        # "provides": {
        #   "type": "object",
        #   "properties": {
        #     ".self": {
        #       "$ref": "#/definitions/resource/properties/spec" #excepting "configurations"
        #      },
        #     ".configurations": {
        #       "allOf": [
        #         { "$ref": "#/definitions/namedObjects" },
        #         # {"additionalProperties": { "$ref": "#/definitions/configurationSpec" }}
        #       ],
        #       'default': {}
        #     },
        #   },
        #   # "additionalProperties": { "$ref": "#/definitions/resource/properties/spec" },
        #   'default': {}
        # },
      },
      "required": ["className"]
    },
    "configurationStatus": {
      "type": "object",
      "allOf": [
        { "$ref": "#/definitions/status" },
        { "properties": {
            "action": { "enum": list(Action.__members__) },
            "inputs": {
              "$ref": "#/definitions/attributes",
              "default": {}
            },
            "modifications": {
              "allOf": [
                { "$ref": "#/definitions/namedObjects" },
                {"additionalProperties": { "$ref": "#/definitions/namedObjects" }}
              ],
              'default': {}
            },
            "dependencies": {
              "type": "array",
              "items": {
                "type": "object",
                "properties": {
                    "name":      { "type": "string" },
                    "ref":       { "type": "string" },
                    "expected":  {},
                    "schema":    { "$ref": "#/definitions/schema" },
                    "required":  { "type": "boolean"},
                },
              },
            },
          }
      }],
    },
    "status": {
      "type": "object",
      "properties": {
        "readyState": {
          "type": "object",
          "properties": {
            "effective": { "enum": list(Status.__members__) },
            "local":    { "enum": list(Status.__members__) }
          }
        },
        "priority": { "enum": list(Priority.__members__) },
        "lastStateChange": {"$ref": "#/definitions/changeId"},
        "lastConfigChange": {"$ref": "#/definitions/changeId"},
      },
      "additionalProperties": True,
    },
    "changeId": {"type": "number"},
    "schema":   {"type": "object"}
  },
  # end definitions
  "type": "object",
  "properties": {
    "apiVersion": { "enum": [ VERSION ] },
    "kind":   { "enum": [ "Manifest" ] },
    "spec":    {"type": "object"},
    "status": {
      "type": "object",
      "allOf": [
        {"properties": {
          "topology": { "type": "string" },
          "inputs": {
            "$ref": "#/definitions/attributes",
           },
          "outputs": {
            "$ref": "#/definitions/attributes",
          },
          "instances": {
            "allOf": [
              { "$ref": "#/definitions/namedObjects" },
              {"additionalProperties": { "$ref": "#/definitions/resource" }}
            ],
          },
        }},
        { "$ref": "#/definitions/status" },
      ],
      'default': {}
    },
    "changes": { "type": "array",
      "additionalItems": {
        "type": "object",
        "allOf": [
          { "$ref": "#/definitions/status" },
          { "$ref": "#/definitions/configurationStatus" },
          { "properties": {
              "parentId": {"$ref": "#/definitions/changeId"},
              'commitId': {"type": "string"},
              'startTime':{"type": "string"},
              "implementation":  {"$ref": "#/definitions/configurationSpec" },
            },
            "required": ["changeId"]
          },
        ]
      }
    }
  },
  "required": ["apiVersion", "kind", "spec"]
}

Import = collections.namedtuple('Import', ['resource', 'spec'])

def saveConfigSpec(spec):
  saved = CommentedMap([
    ("intent", spec.intent.name),
    ("className", spec.className),
  ])
  if spec.majorVersion:
    saved["majorVersion"] = spec.majorVersion
  if spec.minorVersion:
    saved["minorVersion"] = spec.minorVersion
  # if spec.provides:
  #   dotSelf = spec.provides.get('.self')
  #   if dotSelf:
  #     # removed defaults put in by schema
  #     dotSelf.pop('configurations', None)
  #     if not dotSelf.get('attributes'):
  #       dotSelf.pop('attributes', None)
  #   saved["provides"] = spec.provides
  return saved

def saveDependency(dep):
  saved = CommentedMap()
  if dep.name:
    saved['name'] = dep.name
  saved['ref'] = dep.expr
  if dep.expected is not None:
    saved['expected'] = serializeValue(dep.expected)
  if dep.schema is not None:
    saved['schema'] = dep.schema
  if dep.required:
    saved['required'] = dep.required
  if dep.wantList:
    saved['wantList'] = dep.wantList
  return saved

def saveResourceChanges(changes):
  d = CommentedMap()
  for k, v in changes.items():
    d[k] = v[ResourceChanges.attributesIndex] or {}
    if v[ResourceChanges.statusIndex] is not None:
      d[k]['.status'] = v[ResourceChanges.statusIndex].name
    if v[ResourceChanges.addedIndex]:
      d[k]['.added'] = v[ResourceChanges.addedIndex]
  return d

def saveStatus(operational, status=None):
  if status is None:
    status = CommentedMap()
  if not operational.lastChange and operational.status == Status.notapplied:
    # skip status
    return status

  readyState = CommentedMap([
    ("effective", operational.status.name),
  ])
  if operational.localStatus is not None:
    readyState['local'] =  operational.localStatus.name
  status["readyState"] = readyState
  if operational.priority: #and operational.priority != Defaults.shouldRun:
    status["priority"] = operational.priority.name
  if operational.lastStateChange:
    status["lastStateChange"] = operational.lastStateChange
  if operational.lastConfigChange:
    status["lastConfigChange"] = operational.lastConfigChange

  return status

def saveConfigChange(configChange):
  items = [(k, getattr(configChange, k)) for k in ChangeRecordAttributes]
  items.extend( saveStatus(configChange).items() )
  return CommentedMap(items) #CommentedMap so order is preserved in yaml output

def saveResult(value):
  if isinstance(value, collections.Mapping):
    return CommentedMap((key, saveResult(v)) for key, v in value.items())
  elif isinstance(value, (collections.MutableSequence, tuple)):
    return [saveResult(item) for item in value]
  elif value is not None and not isinstance(value, (numbers.Real, bool)):
    return toYamlText(value)
  else:
    return value

def saveTask(task):
  """
  convert dictionary suitable for serializing as yaml
  or creating a Changeset.
  """
  output = CommentedMap()
  output['changeId'] = task.changeId
  output['parentId'] = task.parentId
  if task.target:
    output['target'] = task.target.key
  saveStatus(task, output)
  output['implementation'] = saveConfigSpec(task.configSpec)
  if task.inputs:
    output['inputs'] = serializeValue(task.inputs)
  changes = saveResourceChanges(task._resourceChanges)
  if changes:
    output['changes'] = changes
  if task.messages:
    output['messages'] = task.messages
  dependencies = [saveDependency(val) for val in task.dependencies.values()]
  if dependencies:
    output['dependencies'] = dependencies
  if task.result:
    result = task.result.result
    if result:
      output['result'] = saveResult(result)
  else:
    output['result'] = "skipped"

  return output

# +./ +../ +/
class YamlManifest(Manifest):

  def __init__(self, manifest=None, path=None, validate=True, localEnv=None):
    assert not (localEnv and (manifest or path)) # invalid combination of args
    self.localEnv = localEnv
    self.repo = localEnv and localEnv.instanceRepo
    self.manifest = YamlConfig(manifest, path or localEnv and localEnv.manifestPath,
                                    validate, schema, self.loadHook)
    manifest = self.manifest.expanded
    spec = manifest.get('spec', {})
    super(YamlManifest, self).__init__(spec, self.manifest.path, localEnv)
    assert self.tosca
    status = manifest.get('status', {})

    self.changeLogPath = manifest.get('changeLog')
    if self.changeLogPath:
      fullPath = os.path.join(self.getBaseDir(), self.changeLogPath)
      if os.path.exists(fullPath):
        changelog = load_yaml(fullPath)
        changes = changelog.get('changes', [])
      else:
        logger.warning("missing changelog: %s", fullPath)
        changes = manifest.get('changes', [])
    else:
      changes = manifest.get('changes', [])
      if localEnv:
        # save changes to a separate file if we're in a local environment
        self.changeLogPath = 'changes.yaml'

    self.changeSets = dict( (c.get('changeId', c.get('jobId', 0)), c) for c in changes)
    lastChangeId = self.changeSets and max(self.changeSets.keys()) or 0

    rootResource = self.createTopologyResource(status)
    for name, instance in spec.get('instances', {}).items():
      if not rootResource.findResource(name):
        # XXX like Plan.createResource() parent should be hostedOn target if defined
        self.loadResource(name, instance or {}, parent=rootResource)

    importsSpec = manifest.get('imports', {})
    if localEnv:
      importsSpec.setdefault('local', {})
      importsSpec.setdefault('secret', {})
    rootResource.imports = self.loadImports(importsSpec)
    rootResource.setBaseDir(self.getBaseDir())

    self._ready(rootResource, lastChangeId)

  def getBaseDir(self):
    return self.manifest.getBaseDir()

  def createTopologyResource(self, status):
    """
    If an instance of the toplogy is recorded in status, load it,
    otherwise create a new resource using the the topology as its template
    """
    # XXX use the substitution_mapping (3.8.12) represent the resource
    template = self.tosca.topology
    operational = self.loadStatus(status)
    root = TopologyResource(template, operational)
    for key, val in status.get('instances', {}).items():
      self._createNodeInstance(Resource, key, val, root)
    return root

  def saveNodeInstance(self, resource):
    status = CommentedMap()
    status['template'] = resource.template.getUri()

    # only save the attributes that were set by the instance, not spec properties or attribute defaults
    # particularly, because these will get loaded in later runs and mask any spec properties with the same name
    if resource._attributes:
      status['attributes'] = resource._attributes
    saveStatus(resource, status)
    if resource.createdOn: #will be a ChangeRecord
      status['createdOn'] = resource.createdOn.changeId
    return (resource.name, status)

  def saveResource(self, resource, workDone):
    name, status = self.saveNodeInstance(resource)
    if resource.capabilities:
      status['capabilities'] = CommentedMap(map(self.saveNodeInstance, resource.capabilities))
    if resource.resources:
      status['resources'] = CommentedMap(map(lambda r: self.saveResource(r, workDone), resource.resources))
    return (name, status)

  def saveRootResource(self, workDone):
    resource = self.rootResource
    status = CommentedMap()

    # record the input and output values
    # XXX make sure sensative values are redacted (and need to check for 'sensitive' metadata)
    status['inputs'] = serializeValue(resource.inputs.attributes)
    status['outputs'] = serializeValue(resource.outputs.attributes)

    saveStatus(resource, status)
    # getOperationalDependencies() skips inputs and outputs
    status['instances'] = CommentedMap(map(lambda r: self.saveResource(r, workDone),
                                                resource.getOperationalDependencies()))
    return status

  def saveJobRecord(self, job):
    """
    jobId: 1 # should be last that ran, instead of first?
    startCommit: '' #commitId when job began
    startTime:
    specDigest:
    lastChangeId:
    readyState:
      effective: error
      local: ok
    """
    output = CommentedMap()
    output['jobId'] = job.changeId
    output['startTime'] = job.startTime
    if self.currentCommitId:
      output['startCommit'] = self.currentCommitId
    output['specDigest'] = self.specDigest
    output['lastChangeId'] = job.runner.lastChangeId
    return saveStatus(job, output)

  def saveJob(self, job):
    changed = self.saveRootResource(job.workDone)
    # XXX imported resources need to include its repo's workingdir commitid in their status
    # status and job's changeset also need to save status of repositories
    # that were accessed by loadFromRepo() and add them with commitid and repotype
    # note: initialcommit:requiredcommit means any repo that has at least requiredcommit

    # update changed with includes, this may change objects with references to these objects
    restoreIncludes(self.manifest.includes, self.manifest.config, changed, cls=CommentedMap)
    # modify original to preserve structure and comments
    if 'status' not in self.manifest.config:
      self.manifest.config['status'] = {}

    if not self.manifest.config['status']:
      self.manifest.config['status'] = changed
    else:
      patchDict(self.manifest.config['status'], changed, cls=CommentedMap)

    jobRecord = self.saveJobRecord(job)
    if job.workDone:
      self.manifest.config['latestChange'] = jobRecord
      changes = map(saveTask, job.workDone.values())
      if self.changeLogPath:
        self.manifest.config['changeLog'] = self.changeLogPath
      else:
        self.manifest.config.setdefault('changes', []).extend(changes)
    else:
      # no work was done, so bother recording this job
      changes = []

    if job.out:
      self.dump(job.out)
    else:
      output = six.StringIO()
      self.dump(output)
      job.out = output
      if self.manifest.path:
        with open(self.manifest.path, 'w') as f:
          f.write(output.getvalue())
    return jobRecord, changes

  def dump(self, out=sys.stdout):
    try:
      self.manifest.dump(out)
    except:
      raise UnfurlError("Error saving manifest %s" % self.manifest.path, True)

  def commitJob(self, job):
    if job.planOnly:
      return
    if job.dryRun:
      logger.info("printing results from dry run")
      if not job.out and self.manifest.path:
        job.out = sys.stdout
    jobRecord, changes = self.saveJob(job)
    if not changes:
      logger.info("job run didn't make any changes; nothing to commit")
      return
    if job.dryRun:
      return

    if self.repo:
      self.repo.repo.index.add([self.manifest.path])
      self.repo.repo.index.commit("Updating status for job %s" % job.changeId)
      jobRecord['endCommit'] = self.repo.revision
    if self.changeLogPath:
      self.saveChangeLog(jobRecord, changes)
      if self.repo:
        self.repo.repo.index.add([os.path.join(self.getBaseDir(), self.changeLogPath)])
        self.repo.repo.index.commit("Updating changelog for job %s" % job.changeId)
        logger.info("committed instance repo changes: %s", self.repo.revision)

  def saveChangeLog(self, jobRecord, newChanges):
    """
    manifest: manifest.yaml
    changes:
    """
    try:
      changelog = CommentedMap()
      changelog['manifest'] = os.path.basename(self.manifest.path)
      # put jobs before their child tasks
      key = lambda r: r.get('lastChangeId', r.get('changeId',0))
      changes = itertools.chain([jobRecord], newChanges, self.changeSets.values())
      changelog['changes'] = sorted(changes, key=key, reverse = True)
      output = six.StringIO()
      yaml.dump(changelog, output)
      fullPath = os.path.join(self.getBaseDir(), self.changeLogPath)
      logger.info("saving changelog to %s", fullPath)
      with open(fullPath, 'w') as f:
        f.write(output.getvalue())
    except:
      raise UnfurlError("Error saving changelog %s" % self.changeLogPath, True)

  def loadImports(self, importsSpec):
    """
      file: local/path # for now
      repository: uri or repository name in TOSCA template
      commitId:
      resource: name # default is root
      attributes: # queries into resource
      properties: # expected schema for attributes
    """
    imports = {}
    for name, value in importsSpec.items():
      resource = self.localEnv and self.localEnv.getLocalResource(name, value)
      if not resource:
        if 'file' not in value:
          raise UnfurlError("Can not import '%s': no file specified" % (name))
        # use tosca's loader instead, this can be an url into a repo
        # if repo is url to a git repo, find or create working dir
        repositories = self.tosca.template.tpl.get('repositories',{})
        # load an instance repo
        importDef = value.copy()
        path, yamlDict = loadFromRepo(name, importDef, self.getBaseDir(), repositories, self)
        imported = YamlManifest(yamlDict, path=path)
        rname = value.get('resource', 'root')
        resource = imported.getRootResource().findResource(rname)
        if 'inheritHack' in value:
          value['inheritHack']._attributes['inheritFrom'] = resource
          resource = value.pop('inheritHack')
      if not resource:
        raise UnfurlError("Can not import '%s': resource '%s' not found" % (name, rname))
      imports[name] = Import(resource, value)
    return imports

def runJob(manifestPath=None, _opts=None):
  _opts = _opts or {}
  localEnv = LocalEnv(manifestPath, _opts.get('home'))
  opts = JobOptions(**_opts)
  path = localEnv.manifestPath
  if opts.planOnly:
    logger.info("creating plan for %s", path)
  else:
    logger.info("running job for %s", path)

  logger.info('loading manifest at %s', path)
  try:
    manifest = YamlManifest(localEnv=localEnv)
  except Exception as e:
    logger.error('failed to load manifest at %s: %s', path, str(e),
                                  exc_info=opts.verbose >= 2)
    return None

  runner = Runner(manifest)
  return runner.run(opts)



import unittest
import os
from unfurl.result import ResultsList, serializeValue
from unfurl.eval import Ref, mapValue, RefContext
from unfurl.support import applyTemplate
from unfurl.runtime import Resource
from ruamel.yaml.comments import CommentedMap

class EvalTest(unittest.TestCase):
  longMessage = True

  def test_CommentedMap(self):
    cm = CommentedMap()
    # check bug in ruamel.yaml is fixed: raises TypeError: source has undefined order
    self.assertEqual(cm, cm.copy())

  def test_CommentedMapEquality(self):
    cm = CommentedMap((('b', 2),))
    cm.insert(1, 'a', 1, comment="a comment")
    self.assertEqual(cm, {'a':1, 'b': 2})

  def _getTestResource(self, more=None):
    resourceDef = {
      "name": "test",
      'a': {'ref': 'name'},
      'b': [1, 2, 3],
      'd': {'a':'va', 'b':'vb'},
      'n': {'n':{'n':'n'}},
      's': {'ref': '.'},
      'x': [{
              'a': [{'c':1}, {'c':2}, {'b': 'exists'}, {'l': ['l1']}, {'l': ['l2']}],
            },
            [{'c':5}],
            {
              "a": [{'c':3}, {'c':4}, {'l': ['l3']}, {'l': ['l4']}]
            },
            [{'c':6}],
          ],
      'e': {'a1':{'b1': 'v1'},
            'a2':{'b2': 'v2'}},
      'f': {'a': 1, 'b': {'ref': '.::f::a'} },
      }
    if more:
      resourceDef.update(more)
    resource = Resource("test", attributes=resourceDef)
    assert resource.attributes['x'] == resourceDef['x']
    assert resource.attributes['a'] == 'test'
    assert resource.attributes['s'] is resource
    return resource

  def test_refs(self):
    assert Ref.isRef({'ref': '::name'})
    assert not Ref.isRef({'ref': '::name', 'somethingUnexpected': 1})
    assert Ref.isRef({
      'ref': '::name',
      'vars': {
        'a': None
      }
    })

  def test_refPaths(self):
    resource = self._getTestResource()
    for (exp, expected) in [
      ['x?::a[c=4]', [[{'c':3}, {'c':4}, {'l': ['l3']}, {'l': ['l4']}]] ],
      ['x::a[c]?', [[{'c':1}, {'c':2}, {'b': 'exists'}, {'l': ['l1']}, {'l': ['l2']}]]],
      ['x::a::[c]', [{'c':1}, {'c':2}, {'c':3}, {'c':4}]],
      ['x::a?::[c]', [{'c':1}, {'c':2}]],
      ['a', ['test']],
      ['b', [[1,2,3]]],
      ['b::0', [1]],
      ['b?::2', [3]],
      ['[b::2]::b::2', [3]],
      ['b::1', [2]],
      ['s::b', [[1, 2, 3]]],
      ['s::b::1', [2]],
      ['s::s::b::1', [2]],
      ['n::n::n', ['n']],
      ['d[a=va]', [{'a':'va', 'b':'vb'}]],
      ['d[a=vb]', []],
      ['b[1=2]', [[1,2,3]]],
      ['b[1=1]', []],
      ['a[=test]', ['test']],
      ['a[!=test]', []],
      ['a[key]', []],
      ['d[a=va][b=vb]', [{'a':'va', 'b':'vb'}]],
      ['d[a=va][a=vb]', []],
      ['d[a=va][a!=vb]', [{'a':'va', 'b':'vb'}]],
      ['d[a=va]::b', ['vb']],
      ['x::a::c', [1,2,3,4]],
      ['x::c', [5, 6]],
      ['x::[c]', [{'c':5}, {'c':6}]],
      ['x::a[b]::c', [1,2]],
      ['x::a[!b]::c', [3,4]],
      ['x::a::l', [['l1'],['l2'],['l3'],['l4']]],
      [{'ref': 'a[=$yes]',
        'vars': {'yes': 'test'}
       },  ['test']],
      [{'ref': 'a[=$no]',
        'vars': {'no': None}
       },  []],
       ['[a]', [resource]],
       ['[=blah]', []],
       ['[blah]', []],
       ['[!blah]', [resource]],
       ['.[!=blah]', [resource]],
       ['[!a]', []],
       ['::test', [resource]],
       ['d::*', set(['va', 'vb'])],
       ['e::*::b2', ['v2']],
       ["*", []],
       ["f", [{'a': 1, 'b': 1}]],
       # [{"q": "{{ foo }}"}, ["{{ foo }}"]]
       #XXX test nested ['.[k[d=3]=4]']
    ]:
      ref = Ref(exp)
      # print ('eval', ref.source, ref)
      if isinstance(expected, set):
        # for results where order isn't guaranteed in python2.7
        self.assertEqual(set(ref.resolve(RefContext(resource))), expected, "expr was: " + ref.source)
      else:
        self.assertEqual(ref.resolve(RefContext(resource, trace=0)), expected, "expr was: " + ref.source)

  def test_funcs(self):
    resource = self._getTestResource()
    test1 = {
      'ref': '.name',
      'vars': {
        'a': None
      }
    }
    test2 = {
      'ref': '$b',
      'vars': {
        'b': 1
      }
    }
    test3 = {
      'ref': {
        'if': {'not': '$a'},
        'then': {'q': 'expected'},
        'else': {'q': 'unexpected'},
      },
      'vars': {
        'a': None
      }
    }
    result1 = Ref(test1).resolveOne(RefContext(resource))
    self.assertEqual('test', result1)
    result2 = Ref(test2).resolveOne(RefContext(resource))
    self.assertEqual(1, result2)
    result3 = Ref(test3).resolveOne(RefContext(resource))
    self.assertEqual('expected', result3)
    result4 = Ref(test3).resolve(RefContext(resource))
    self.assertEqual(['expected'], result4)
    test5 = {
      'ref': {
        'or': ['$a', 'b'],
      },
      'vars': {
        'a': None
      }
    }
    result5 = Ref(test5).resolveOne(RefContext(resource))
    self.assertEqual(resource.attributes['b'], result5) # this doesn't seem obvious!

  def test_forEach(self):
    resource = self._getTestResource()
    test1 = {
      'ref': '.',
      'foreach': {
          'value': {'content': { 'ref': 'b'}}
      },
    }
    expected0 = {'content': [1, 2, 3]}
    expected = {
      'UmVzb3VyY2UoJ3Rlc3QnKQ==': expected0
    }
    result0 = Ref(test1).resolveOne(RefContext(resource, trace=0))
    self.assertEqual(expected0, result0)
    # resolve has same result as resolveOne
    self.assertEqual([expected0], Ref(test1).resolve(RefContext(resource)))

    #add 'key' to make result a dict
    # test that template strings work
    test1['foreach']['key'] = '{{ item | ref | b64encode}}';
    result1 = Ref(test1).resolveOne(RefContext(resource))
    self.assertEqual(expected, result1)
    result2 = Ref(test1).resolve(RefContext(resource))
    self.assertEqual([expected], result2)

  def test_serializeValues(self):
    resource = self._getTestResource()
    src = {'a': ['b', resource]}
    serialized = serializeValue(src)
    self.assertEqual(serialized, {'a': ['b', {'ref': '::test'}]})
    self.assertEqual(src, mapValue(serialized, resource))

  def test_jinjaTemplate(self):
    resource = Resource("test", attributes=dict(a1="hello"))
    ctx = RefContext(resource, {"foo": "hello"})

    self.assertEqual(applyTemplate(" {{ foo }} ", ctx), "hello")
    # test jinja2 native types
    self.assertEqual(applyTemplate(" {{[foo]}} ", ctx), ["hello"])

    self.assertEqual(applyTemplate(' {{ "::test::a1" | ref }} ', ctx), u"hello")
    self.assertEqual(applyTemplate(' {{ lookup("unfurl", "::test::a1") }} ', ctx), u"hello")
    # ansible query() always returns a list
    self.assertEqual(applyTemplate('{{  query("unfurl", "::test::a1") }}', ctx), [u'hello'])

    os.environ['TEST_ENV'] = 'testEnv' # note: tox doesn't pass on environment variables so we need to set one now
    self.assertEqual(mapValue("{{ lookup('env', 'TEST_ENV') }}", resource), 'testEnv')

    #test that ref vars as can be used as template string vars
    exp = {'a': "{{ aVar }} world"}
    vars = {'aVar': 'hello'}
    self.assertEqual(mapValue(exp, RefContext(resource, vars)), {'a': 'hello world'})

  def test_templateFunc(self):
    query = {
      "eval": {"template": "{%if testVar %}{{success}}{%else%}failed{%endif%}"},
      "vars": {"testVar": True, "success": dict(eval={
        'if': '$true',
        'then': '.name'})
      }
    }
    resource = self._getTestResource({'aTemplate': query})
    self.assertEqual(mapValue(query, resource), 'test')
    self.assertEqual(resource.attributes['aTemplate'], 'test')

  def test_innerReferences(self):
    resourceDef = {'a':
      dict(b = {"ref": "a::c"},
        c= {'e': 1},
        d= ['2', {"ref": "a::d::0"}]
      )
    }
    resource = Resource("test", attributes=resourceDef)
    assert not not resource.attributes
    self.assertEqual(len(resource.attributes), 1)

    expectedA = {'c': {'e': 1}, 'b': {'e': 1}, 'd': ['2', '2']}
    self.assertEqual(resource.attributes['a']['b'], expectedA['b'])
    self.assertEqual(resource.attributes['a'], expectedA)
    self.assertEqual(Ref("a").resolve(RefContext(resource)), [expectedA])
    self.assertEqual(Ref("a").resolveOne(RefContext(resource)), expectedA)

    expected = ['2']
    self.assertEqual(Ref("a::d::0").resolve(RefContext(resource)), expected)
    self.assertEqual(Ref("a::d::1").resolve(RefContext(resource)), expected)

    # print('test_references', resource.attributes,
    #   'AAAA', resource.attributes['a'],
    #   'BBB', resource.attributes['a']['b'],
    # )
    self.assertEqual(resource.attributes['a'], expectedA)
    self.assertEqual(resource.attributes['a']['d'][0], '2')
    self.assertEqual(resource.attributes['a']['d'][1], '2')
    self.assertEqual(resource.attributes['a']['b']['e'], 1)

    self.assertEqual(Ref("a::b::e").resolve(RefContext(resource)), [1])

    # test again to make sure it still resolves correctly
    self.assertEqual(Ref("a::d::0").resolve(RefContext(resource)), expected)
    self.assertEqual(Ref("a::d::1").resolve(RefContext(resource)), expected)

  def test_vars(self):
    # test dereferencing vars
    resource = self._getTestResource()
    query = {
      'eval': "$aDict",
      'vars': {
        "aDict": {
            'aRef': {"eval": '::test'},
            "aTemplate": "{{ true }}"
        }
      }
    }
    result = Ref(query).resolveOne(RefContext(resource))
    self.assertEqual(result, {'aRef': resource, 'aTemplate': True})

    query = {
      'eval': "$aRef",
      'vars': {
        "aRef": {"eval": '::test'}
      }
    }
    assert Ref.isRef(query['vars']['aRef'])
    result = Ref(query).resolveOne(RefContext(resource))
    self.assertEqual(result, resource)

  def test_nodeTraversal1(self):
    root = Resource('r2', { "a": [
            dict(ref="::r1::a"), #'r1'
            dict(ref="b") #'r2'
          ],
        "b" : 'r2'
    })
    child = Resource('r1', {'a': dict(ref="b"), 'b': 'r1'}, root)
    ctx = RefContext(root)
    x = [{
      'a': [{'c':1}, {'c':2}],
    }]
    assert x == ResultsList(x, ctx)
    self.assertEqual(Ref('b').resolve(RefContext(child)), ['r1'])
    self.assertEqual(Ref('a').resolve(RefContext(child)), ['r1'])
    self.assertEqual(Ref('a').resolve(RefContext(root)), [['r1', 'r2']])

  def test_nodeTraversal2(self):
    root = Resource('root', {
      'a': [{'ref':'::child'}, {'b': 2}]
    })
    child = Resource('child', {'b': 1}, root)
    self.assertEqual(Ref('.ancestors').resolve(RefContext(child)), [[child, root]])
    # self.assertEqual(Ref('a::b').resolve(RefContext(root)), [1])
    self.assertEqual(Ref('a').resolve(RefContext(child)), [[child, {'b': 2}]])
    # a resolves to [child, dict] so a::b resolves to [child[b], [b]2]
    self.assertEqual(Ref('a::b').resolve(RefContext(child)), [1, 2])

  def test_lookup(self):
    resource = self._getTestResource()
    os.environ['TEST_ENV'] = 'testEnv' # note: tox doesn't pass on environment variables so we need to set one now
    query = {
      "eval": {"lookup": {"env": 'TEST_ENV'}}
    }
    self.assertEqual(mapValue(query, resource), 'testEnv')
